# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import difflib
import io
import os
import socket
from unittest import mock

import netaddr
from neutron.agent.l3 import dvr_edge_router
from neutron.agent.l3 import dvr_snat_ns
from neutron.agent.l3 import legacy_router
from neutron.agent.linux import iptables_manager
from neutron.conf.agent.l3 import config as l3_config
from neutron_lib import constants
from neutron_lib.exceptions import vpn as vpn_exception
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_vpnaas.services.vpn.device_drivers import ipsec as openswan_ipsec
from neutron_vpnaas.services.vpn.device_drivers import libreswan_ipsec
from neutron_vpnaas.services.vpn.device_drivers import strongswan_ipsec
from neutron_vpnaas.tests import base

# Note: process_id == router_id == vpnservice_id

_uuid = uuidutils.generate_uuid
FAKE_UUID = _uuid()
FAKE_HOST = 'fake_host'
FAKE_ROUTER_ID = FAKE_UUID
FAKE_VPNSERVICE_ID = FAKE_UUID
FAKE_PROCESS_ID = FAKE_UUID
FAKE_IPSEC_SITE_CONNECTION1_ID = _uuid()
FAKE_IPSEC_SITE_CONNECTION2_ID = _uuid()
FAKE_IKE_POLICY = {
    'ike_version': 'v1',
    'encryption_algorithm': 'aes-128',
    'auth_algorithm': 'sha1',
    'pfs': 'group5',
    'lifetime_value': 3600
}

FAKE_IPSEC_POLICY = {
    'encryption_algorithm': 'aes-128',
    'auth_algorithm': 'sha1',
    'pfs': 'group5',
    'transform_protocol': 'esp',
    'lifetime_value': 3600,
    'encapsulation_mode': 'tunnel'
}

FAKE_VPN_SERVICE = {
    'id': FAKE_VPNSERVICE_ID,
    'router_id': FAKE_ROUTER_ID,
    'name': 'myvpn',
    'admin_state_up': True,
    'status': constants.PENDING_CREATE,
    'external_ip': '60.0.0.4',
    'ipsec_site_connections': [
        {'peer_cidrs': ['20.0.0.0/24',
                        '30.0.0.0/24'],
         'local_cidrs': ['10.0.0.0/24'],
         'local_ip_vers': 4,
         'admin_state_up': True,
         'id': FAKE_IPSEC_SITE_CONNECTION1_ID,
         'external_ip': '60.0.0.4',
         'local_id': '60.0.0.4',
         'peer_address': '60.0.0.5',
         'mtu': 1500,
         'peer_id': '60.0.0.5',
         'psk': 'password',
         'initiator': 'bi-directional',
         'ikepolicy': FAKE_IKE_POLICY,
         'ipsecpolicy': FAKE_IPSEC_POLICY,
         'dpd_action': 'hold',
         'dpd_interval': 30,
         'dpd_timeout': 120,
         'status': constants.PENDING_CREATE},
        {'peer_cidrs': ['40.0.0.0/24',
                        '50.0.0.0/24'],
         'local_cidrs': ['11.0.0.0/24'],
         'local_ip_vers': 4,
         'admin_state_up': True,
         'external_ip': '60.0.0.4',
         'local_id': '60.0.0.4',
         'peer_address': '60.0.0.6',
         'peer_id': '60.0.0.6',
         'mtu': 1500,
         'psk': 'password',
         'id': FAKE_IPSEC_SITE_CONNECTION2_ID,
         'initiator': 'bi-directional',
         'ikepolicy': FAKE_IKE_POLICY,
         'ipsecpolicy': FAKE_IPSEC_POLICY,
         'dpd_action': 'hold',
         'dpd_interval': 30,
         'dpd_timeout': 120,
         'status': constants.PENDING_CREATE}]
}

AUTH_ESP = '''esp
    # [encryption_algorithm]-[auth_algorithm];[pfs]
    phase2alg=aes128-sha1;modp1536'''

AUTH_AH = '''ah
    # AH protocol does not support encryption
    # [auth_algorithm];[pfs]
    phase2alg=sha1;modp1536'''

OPENSWAN_CONNECTION_DETAILS = '''# rightsubnet=networkA/netmaskA, networkB/netmaskB (IKEv2 only)
    # [mtu]
    mtu=1500
    # [dpd_action]
    dpdaction=%(dpd_action)s
    # [dpd_interval]
    dpddelay=%(dpd_delay)s
    # [dpd_timeout]
    dpdtimeout=%(dpd_timeout)s
    # [auth_mode]
    authby=secret
    ######################
    # IKEPolicy params
    ######################
    #ike version
    ikev2=never
    # [encryption_algorithm]-[auth_algorithm];[pfs]
    ike=aes128-sha1;modp1536
    # [lifetime_value]
    ikelifetime=%(ike_lifetime)ss
    # NOTE: it looks lifetime_units=kilobytes can't be enforced \
(could be seconds,  hours,  days...)
    ##########################
    # IPsecPolicys params
    ##########################
    # [transform_protocol]
    phase2=%(auth_mode)s
    # [encapsulation_mode]
    type=%(encapsulation_mode)s
    # [lifetime_value]
    lifetime=%(life_time)ss
    # lifebytes=100000 if lifetime_units=kilobytes (IKEv2 only)
'''  # noqa: E501

IPV4_NEXT_HOP = '''# NOTE: a default route is required for %defaultroute to work...
    leftnexthop=%defaultroute
    rightnexthop=%defaultroute'''  # noqa: E501

IPV6_NEXT_HOP = '''# To recognize the given IP addresses in this config
    # as IPv6 addresses by pluto whack. Default is ipv4
    connaddrfamily=ipv6
    # openswan can't process defaultroute for ipv6.
    # Assign gateway address as leftnexthop
    leftnexthop=%s
    # rightnexthop is not mandatory for ipsec, so no need in ipv6.'''

EXPECTED_OPENSWAN_CONF = """
# Configuration for %(vpnservice_id)s
config setup
    nat_traversal=yes
    virtual_private=%(virtual_privates)s
conn %%default
    keylife=60m
    keyingtries=%%forever
conn %(conn1_id)s
    %(next_hop)s
    left=%(left)s
    leftid=%(leftid)s
    auto=start
    # NOTE:REQUIRED
    # [subnet]
    leftsubnet%(local_cidrs1)s
    # [updown]
    # What "updown" script to run to adjust routing and/or firewalling when
    # the status of the connection changes (default "ipsec _updown").
    # "--route yes" allows to specify such routing options as mtu and metric.
    leftupdown="ipsec _updown --route yes"
    ######################
    # ipsec_site_connections
    ######################
    # [peer_address]
    right=%(right1)s
    # [peer_id]
    rightid=%(right1)s
    # [peer_cidrs]
    rightsubnets={ %(peer_cidrs1)s }
    %(conn_details)sconn %(conn2_id)s
    %(next_hop)s
    left=%(left)s
    leftid=%(leftid)s
    auto=start
    # NOTE:REQUIRED
    # [subnet]
    leftsubnet%(local_cidrs2)s
    # [updown]
    # What "updown" script to run to adjust routing and/or firewalling when
    # the status of the connection changes (default "ipsec _updown").
    # "--route yes" allows to specify such routing options as mtu and metric.
    leftupdown="ipsec _updown --route yes"
    ######################
    # ipsec_site_connections
    ######################
    # [peer_address]
    right=%(right2)s
    # [peer_id]
    rightid=%(right2)s
    # [peer_cidrs]
    rightsubnets={ %(peer_cidrs2)s }
    %(conn_details)s
"""

STRONGSWAN_AUTH_ESP = 'esp=aes128-sha1-modp1536'

STRONGSWAN_AUTH_AH = 'ah=sha1-modp1536'

EXPECTED_IPSEC_OPENSWAN_SECRET_CONF = '''
# Configuration for %s
60.0.0.4 60.0.0.5 : PSK 0scGFzc3dvcmQ=
60.0.0.4 60.0.0.6 : PSK 0scGFzc3dvcmQ=''' % FAKE_VPNSERVICE_ID

EXPECTED_IPSEC_STRONGSWAN_CONF = '''
# Configuration for %(vpnservice_id)s
config setup

conn %%default
        keylife=20m
        rekeymargin=3m
        keyingtries=1
        authby=psk
        mobike=no

conn %(conn1_id)s
    keyexchange=ikev1
    left=%(left)s
    leftsubnet=%(local_cidrs1)s
    leftid=%(leftid)s
    leftfirewall=yes
    right=%(right1)s
    rightsubnet=%(peer_cidrs1)s
    rightid=%(right1)s
    auto=route
    dpdaction=%(dpd_action)s
    dpddelay=%(dpd_delay)ss
    dpdtimeout=%(dpd_timeout)ss
    ike=%(ike_encryption_algorithm)s-%(ike_auth_algorithm)s-%(ike_pfs)s
    ikelifetime=%(ike_lifetime)ss
    %(auth_mode)s
    lifetime=%(life_time)ss
    type=%(encapsulation_mode)s

conn %(conn2_id)s
    keyexchange=ikev1
    left=%(left)s
    leftsubnet=%(local_cidrs2)s
    leftid=%(leftid)s
    leftfirewall=yes
    right=%(right2)s
    rightsubnet=%(peer_cidrs2)s
    rightid=%(right2)s
    auto=route
    dpdaction=%(dpd_action)s
    dpddelay=%(dpd_delay)ss
    dpdtimeout=%(dpd_timeout)ss
    ike=%(ike_encryption_algorithm)s-%(ike_auth_algorithm)s-%(ike_pfs)s
    ikelifetime=%(ike_lifetime)ss
    %(auth_mode)s
    lifetime=%(life_time)ss
    type=%(encapsulation_mode)s
'''

EXPECTED_STRONGSWAN_DEFAULT_CONF = '''
charon {
        load_modular = yes
        plugins {
                include strongswan.d/charon/*.conf
        }
}

include strongswan.d/*.conf
'''

EXPECTED_IPSEC_STRONGSWAN_SECRET_CONF = '''
# Configuration for %s
60.0.0.4 60.0.0.5 : PSK 0scGFzc3dvcmQ=

60.0.0.4 60.0.0.6 : PSK 0scGFzc3dvcmQ=
''' % FAKE_VPNSERVICE_ID

PLUTO_ACTIVE_STATUS = """000 "{conn_id}/0x1": erouted;\n
000 #4: "{conn_id}/0x1":500 STATE_QUICK_R2 (IPsec SA established); \
newest IPSEC;""".format(
    conn_id=FAKE_IPSEC_SITE_CONNECTION2_ID)
PLUTO_ACTIVE_STATUS_IKEV2 = """000 "{conn_id}/0x1": erouted;\n
000 #4: "{conn_id}/0x1":500 STATE_PARENT_R2 (PARENT SA established); \
newest IPSEC;""".format(
    conn_id=FAKE_IPSEC_SITE_CONNECTION2_ID)
PLUTO_MULTIPLE_SUBNETS_ESTABLISHED_STATUS = """000 "{conn_id1}/1x1": erouted;\n
000 #4: "{conn_id1}/1x1":500 STATE_QUICK_R2 (IPsec SA established); \
newest IPSEC;\n
000 "{conn_id2}/2x1": erouted;\n
000 #4: "{conn_id2}/2x1":500 STATE_QUICK_R2 (IPsec SA established); \
newest IPSEC;\n""".format(  # noqa: E501
    conn_id1=FAKE_IPSEC_SITE_CONNECTION1_ID,
    conn_id2=FAKE_IPSEC_SITE_CONNECTION2_ID)
PLUTO_ACTIVE_NO_IPSEC_SA_STATUS = """000 "{conn_id}/0x1": erouted;\n
000 #258: "{conn_id}/0x1":500 STATE_MAIN_R2
(sent MR2, expecting MI3);""".format(
    conn_id=FAKE_IPSEC_SITE_CONNECTION2_ID)
PLUTO_DOWN_STATUS = "000 \"%(conn_id)s/0x1\": unrouted;" % {
    'conn_id': FAKE_IPSEC_SITE_CONNECTION2_ID}

CHARON_ACTIVE_STATUS = "%(conn_id)s{1}:  INSTALLED, TUNNEL" % {
    'conn_id': FAKE_IPSEC_SITE_CONNECTION2_ID}
CHARON_DOWN_STATUS = "%(conn_id)s{1}:  ROUTED, TUNNEL" % {
    'conn_id': FAKE_IPSEC_SITE_CONNECTION2_ID}

NOT_RUNNING_STATUS = "Command: ['ipsec', 'status'] Exit code: 3 Stdout:"


class BaseIPsecDeviceDriver(base.BaseTestCase):
    def setUp(self, driver=openswan_ipsec.OpenSwanDriver,
              ipsec_process=openswan_ipsec.OpenSwanProcess,
              vpnservice=FAKE_VPN_SERVICE):
        super().setUp()
        for klass in [
            'neutron_lib.rpc.Connection',
            'oslo_service.loopingcall.FixedIntervalLoopingCall'
        ]:
            mock.patch(klass).start()
        self._execute = mock.patch.object(ipsec_process, '_execute').start()
        self.agent = mock.Mock()
        self.conf = cfg.CONF
        l3_config.register_l3_agent_config_opts(l3_config.OPTS, self.conf)
        self.agent.conf = self.conf
        self.driver = driver(
            self.agent,
            FAKE_HOST)
        self.driver.agent_rpc = mock.Mock()
        self.ri_kwargs = {'router': {'id': FAKE_ROUTER_ID, 'ha': False},
                          'agent_conf': self.conf,
                          'interface_driver': mock.sentinel.interface_driver}
        self.iptables = mock.Mock()
        self.apply_mock = mock.Mock()
        self.vpnservice = copy.deepcopy(vpnservice)
        ipsec_process._get_strongswan_piddir = mock.Mock(
            return_value="/var/run")

    @staticmethod
    def generate_diff(a, b):
        """Generates unified diff of a and b."""
        a, b = list(a.splitlines(True)), list(b.splitlines(True))
        diff = difflib.unified_diff(a, b, fromfile="expected",
                                    tofile="actual")
        return diff

    def modify_config_for_test(self, overrides):
        """Revise service/connection settings to test variations.

        Must update service, so that dialect mappings occur for any changes
        that are made.
        """
        ipsec_auth_protocol = overrides.get('ipsec_auth')
        if ipsec_auth_protocol:
            auth_proto = {'transform_protocol': ipsec_auth_protocol}
            for conn in self.vpnservice['ipsec_site_connections']:
                conn['ipsecpolicy'].update(auth_proto)

        local_cidrs = overrides.get('local_cidrs')
        if local_cidrs:
            for i, conn in enumerate(
                    self.vpnservice['ipsec_site_connections']):
                conn['local_cidrs'] = local_cidrs[i]

        local_ip_version = overrides.get('local_ip_vers', 4)
        for conn in self.vpnservice['ipsec_site_connections']:
            conn['local_ip_vers'] = local_ip_version

        peer_cidrs = overrides.get('peer_cidrs')
        if peer_cidrs:
            for i, conn in enumerate(
                    self.vpnservice['ipsec_site_connections']):
                conn['peer_cidrs'] = peer_cidrs[i]

        peers = overrides.get('peers')
        if peers:
            for i, conn in enumerate(
                    self.vpnservice['ipsec_site_connections']):
                conn['peer_id'] = peers[i]
                conn['peer_address'] = peers[i]

        local_ip = overrides.get('local')
        local_id = overrides.get('local_id')
        if local_ip:
            for conn in self.vpnservice['ipsec_site_connections']:
                conn['external_ip'] = local_ip
                conn['local_id'] = local_ip
                if local_id:
                    conn['local_id'] = local_id

    def check_config_file(self, expected, actual):
        expected = expected.strip()
        actual = actual.strip()
        res_diff = self.generate_diff(expected, actual)
        self.assertEqual(expected, actual, message=''.join(res_diff))

    def _test_ipsec_connection_config(self, info):
        """Check config file string for service/connection.

        Calls test specific method to create (and override as needed) the
        expected config file string, generates the config using the test's
        IPSec template, and then compares the results.
        """

        expected = self.build_ipsec_expected_config_for_test(info)
        actual = self.process._gen_config_content(self.ipsec_template,
                                                  self.vpnservice)
        self.check_config_file(expected, actual)


class IPSecDeviceLegacy(BaseIPsecDeviceDriver):

    def setUp(self, driver=openswan_ipsec.OpenSwanDriver,
              ipsec_process=openswan_ipsec.OpenSwanProcess):
        super().setUp(driver, ipsec_process)
        self._make_router_info_for_test()

    def _make_router_info_for_test(self):
        self.router_info = legacy_router.LegacyRouter(router_id=FAKE_ROUTER_ID,
                                                      agent=self.agent,
                                                      **self.ri_kwargs)
        self.router_info.router['distributed'] = False
        self.router_info.iptables_manager.ipv4['nat'] = self.iptables
        self.router_info.iptables_manager.apply = self.apply_mock
        self.driver.routers[FAKE_ROUTER_ID] = self.router_info

    def _test_vpnservice_updated(self, expected_param, **kwargs):
        with mock.patch.object(self.driver, 'sync') as sync:
            context = mock.Mock()
            self.driver.vpnservice_updated(context, **kwargs)
            sync.assert_called_once_with(context, expected_param)

    def test_vpnservice_updated(self):
        self._test_vpnservice_updated([])

    def test_vpnservice_updated_with_router_info(self):
        kwargs = {'router': self.router_info}
        self._test_vpnservice_updated([self.router_info], **kwargs)

    def test_create_router(self):
        process = mock.Mock(openswan_ipsec.OpenSwanProcess)
        process.vpnservice = self.vpnservice
        self.driver.processes = {
            FAKE_ROUTER_ID: process}
        self.driver.create_router(self.router_info)
        self._test_ensure_nat_rules()
        process.enable.assert_called_once_with()

    def test_destroy_router(self):
        process_id = _uuid()
        process = mock.Mock()
        process.vpnservice = self.vpnservice
        self.driver.processes = {
            process_id: process}
        self.driver.destroy_router(process_id)
        process.disable.assert_called_once_with()
        self.assertNotIn(process_id, self.driver.processes)

    def _test_ensure_nat_rules(self):
        self.router_info.iptables_manager.ipv4['nat'].assert_has_calls([
            mock.call.clear_rules_by_tag('vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 10.0.0.0/24 -d 20.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 10.0.0.0/24 -d 30.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 11.0.0.0/24 -d 40.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 11.0.0.0/24 -d 50.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas')
        ])
        self.router_info.iptables_manager.apply.assert_called_once_with()

    def _test_ensure_nat_rules_with_multiple_locals(self):
        self.router_info.iptables_manager.ipv4['nat'].assert_has_calls([
            mock.call.clear_rules_by_tag('vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 10.0.0.0/24 -d 20.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 10.0.0.0/24 -d 30.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 11.0.0.0/24 -d 20.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 11.0.0.0/24 -d 30.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 12.0.0.0/24 -d 40.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 12.0.0.0/24 -d 50.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 13.0.0.0/24 -d 40.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas'),
            mock.call.add_rule(
                'POSTROUTING',
                '-s 13.0.0.0/24 -d 50.0.0.0/24 -m policy '
                '--dir out --pol ipsec -j ACCEPT ',
                top=True,
                tag='vpnaas')
        ])
        self.router_info.iptables_manager.apply.assert_called_once_with()

    def test_sync(self):
        fake_vpn_service = FAKE_VPN_SERVICE
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
            fake_vpn_service]
        context = mock.Mock()
        self.driver._sync_vpn_processes = mock.Mock()
        self.driver._delete_vpn_processes = mock.Mock()
        self.driver._cleanup_stale_vpn_processes = mock.Mock()
        sync_router_ids = [fake_vpn_service['router_id']]
        self.driver.sync(context, [self.router_info])
        self.driver._sync_vpn_processes.assert_called_once_with(
            [fake_vpn_service], sync_router_ids)
        self.driver._delete_vpn_processes.assert_called_once_with(
            sync_router_ids, sync_router_ids)
        self.driver._cleanup_stale_vpn_processes.assert_called_once_with(
            sync_router_ids)

    def test__sync_vpn_processes_new_vpn_service(self):
        new_vpnservice = self.vpnservice
        router_id = new_vpnservice['router_id']
        self.driver.processes = {}
        with mock.patch.object(self.driver, 'ensure_process') as ensure_p:
            ensure_p.side_effect = self.fake_ensure_process
            self.driver._sync_vpn_processes([new_vpnservice], router_id)
            self._test_ensure_nat_rules()
            self.driver.processes[router_id].update.assert_called_once_with()

    def test_ensure_nat_rules_with_multiple_local_subnets(self):
        """Ensure that add nat rule combinations are correct."""
        overrides = {'local_cidrs': [['10.0.0.0/24', '11.0.0.0/24'],
                                     ['12.0.0.0/24', '13.0.0.0/24']]}
        self.modify_config_for_test(overrides)
        self.driver.ensure_nat_rules(self.vpnservice)
        self._test_ensure_nat_rules_with_multiple_locals()

    def test__sync_vpn_processes_router_with_no_vpn(self):
        """Test _sync_vpn_processes with a router not hosting vpnservice.

        This test case tests that when a router which doesn't host
        vpn services is updated, sync_vpn_processes doesn't restart/update
        the existing vpnservice processes.
        """
        process = mock.Mock()
        process.vpnservice = self.vpnservice
        process.connection_status = {}
        self.driver.processes = {
            self.vpnservice['router_id']: process}
        router_id_no_vpn = _uuid()
        with mock.patch.object(self.driver, 'ensure_process') as ensure_p:
            self.driver._sync_vpn_processes([self.vpnservice],
                                            [router_id_no_vpn])
            self.assertEqual(0, ensure_p.call_count)

    def test__sync_vpn_processes_router_with_no_vpn_and_no_vpn_services(self):
        """No vpn services running and router not hosting vpn svc."""
        router_id_no_vpn = _uuid()
        self.driver.process_status_cache = {}
        self.driver.processes = {}
        with mock.patch.object(self.driver, 'ensure_process') as ensure_p:
            ensure_p.side_effect = self.fake_ensure_process
            self.driver._sync_vpn_processes([], [router_id_no_vpn])
            self.assertEqual(0, ensure_p.call_count)

    def test__sync_vpn_processes_router_with_no_vpn_agent_restarted(self):
        """Test for the router not hosting vpnservice and agent restarted.

        This test case tests that when a non vpnservice hosted router
        is updated, _sync_vpn_processes restart/update the existing vpnservices
        which are not yet stored in driver.processes.
        """
        self.driver.process_status_cache = {}
        self.driver.processes = {}
        with mock.patch.object(self.driver, 'ensure_process') as ensure_p:
            ensure_p.side_effect = self.fake_ensure_process
            self.driver._sync_vpn_processes(
                [self.vpnservice],
                [FAKE_ROUTER_ID]
            )
            self._test_ensure_nat_rules()
            self.driver.processes[
                FAKE_ROUTER_ID
            ].update.assert_called_once_with()

    def test_delete_vpn_processes(self):
        router_id_no_vpn = _uuid()
        vpn_service_router_id = _uuid()
        with mock.patch.object(self.driver,
                               'destroy_process') as (fake_destroy_process):
            self.driver._delete_vpn_processes([router_id_no_vpn],
                                              [vpn_service_router_id])
            fake_destroy_process.assert_has_calls(
                [mock.call(router_id_no_vpn)])

        # test that _delete_vpn_processes doesn't delete the
        # the valid vpn processes
        with mock.patch.object(self.driver,
                               'destroy_process') as fake_destroy_process:
            self.driver._delete_vpn_processes([vpn_service_router_id],
                                              [vpn_service_router_id])
            self.assertFalse(fake_destroy_process.called)

    def test_cleanup_stale_vpn_processes(self):
        stale_vpn_service = {'router_id': _uuid()}
        active_vpn_service = {'router_id': _uuid()}
        self.driver.processes = {
            stale_vpn_service['router_id']: stale_vpn_service,
            active_vpn_service['router_id']: active_vpn_service}
        with mock.patch.object(self.driver, 'destroy_process') as destroy_p:
            self.driver._cleanup_stale_vpn_processes(
                [active_vpn_service['router_id']])
            destroy_p.assert_has_calls(
                [mock.call(stale_vpn_service['router_id'])])

    def fake_ensure_process(self, process_id, vpnservice=None):
        process = self.driver.processes.get(process_id)
        if not process:
            process = mock.Mock()
            process.vpnservice = self.vpnservice
            process.connection_status = {}
            process.status = constants.ACTIVE
            process.updated_pending_status = True
            self.driver.processes[process_id] = process
        elif vpnservice:
            process.vpnservice = vpnservice
            process.update_vpnservice(vpnservice)
        return process

    def fake_destroy_router(self, process_id):
        process = self.driver.processes.get(process_id)
        if process:
            del self.driver.processes[process_id]

    # TODO(crohmann): Add test cases for HARouter and different ha_states
    # @ddt [(False, None),(True, 'primary'), (True, 'standby')]
    def test_sync_update_vpnservice(self):
        with mock.patch.object(self.driver,
                               'ensure_process') as ensure_process:
            ensure_process.side_effect = self.fake_ensure_process
            new_vpn_service = self.vpnservice
            updated_vpn_service = copy.deepcopy(new_vpn_service)
            updated_vpn_service['ipsec_site_connections'][1].update(
                {'peer_cidrs': ['60.0.0.0/24', '70.0.0.0/24']})
            context = mock.Mock()
            self.driver.process_status_cache = {}
            self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
                new_vpn_service]
            self.driver.sync(context, [self.router_info])
            process = self.driver.processes[FAKE_ROUTER_ID]
            self.assertEqual(new_vpn_service, process.vpnservice)
            self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
                updated_vpn_service]
            self.driver.sync(context, [self.router_info])
            process = self.driver.processes[FAKE_ROUTER_ID]
            process.update_vpnservice.assert_called_once_with(
                updated_vpn_service)
            self.assertEqual(updated_vpn_service, process.vpnservice)

    def test_sync_removed(self):
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = []
        context = mock.Mock()
        process_id = _uuid()
        process = mock.Mock()
        process.vpnservice = self.vpnservice
        self.driver.processes = {
            process_id: process}
        self.driver.sync(context, [])
        process.disable.assert_called_once_with()
        self.assertNotIn(process_id, self.driver.processes)

    def test_sync_removed_router(self):
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = []
        context = mock.Mock()
        process_id = _uuid()
        ri = self.router_info
        ri.router_id = process_id
        ri.router['id'] = process_id
        self.driver.sync(context, [self.router_info])
        self.assertNotIn(process_id, self.driver.processes)

    def test_status_updated_on_connection_admin_down(self):
        self.driver.process_status_cache = {
            '1': {
                'status': constants.ACTIVE,
                'id': 123,
                'updated_pending_status': False,
                'ipsec_site_connections': {
                    '10': {
                        'status': constants.ACTIVE,
                        'updated_pending_status': False,
                    },
                    '20': {
                        'status': constants.ACTIVE,
                        'updated_pending_status': False,
                    }
                }
            }
        }
        # Simulate that there is no longer status for connection '20'
        # e.g. connection admin down
        new_status = {
            'ipsec_site_connections': {
                '10': {
                    'status': constants.ACTIVE,
                    'updated_pending_status': False
                }
            }
        }
        self.driver.update_downed_connections('1', new_status)
        existing_conn = new_status['ipsec_site_connections'].get('10')
        self.assertIsNotNone(existing_conn)
        self.assertEqual(constants.ACTIVE, existing_conn['status'])
        missing_conn = new_status['ipsec_site_connections'].get('20')
        self.assertIsNotNone(missing_conn)
        self.assertEqual(constants.DOWN, missing_conn['status'])

    def test_status_updated_on_service_admin_down(self):
        self.driver.process_status_cache = {
            '1': {
                'status': constants.ACTIVE,
                'id': 123,
                'updated_pending_status': False,
                'ipsec_site_connections': {
                    '10': {
                        'status': constants.ACTIVE,
                        'updated_pending_status': False,
                    },
                    '20': {
                        'status': constants.ACTIVE,
                        'updated_pending_status': False,
                    }
                }
            }
        }
        # Simulate that there are no connections now
        new_status = {
            'ipsec_site_connections': {}
        }
        self.driver.update_downed_connections('1', new_status)
        missing_conn = new_status['ipsec_site_connections'].get('10')
        self.assertIsNotNone(missing_conn)
        self.assertEqual(constants.DOWN, missing_conn['status'])
        missing_conn = new_status['ipsec_site_connections'].get('20')
        self.assertIsNotNone(missing_conn)
        self.assertEqual(constants.DOWN, missing_conn['status'])

    def _test_status_handling_for_downed_connection(self, down_status):
        """Test status handling for downed connection."""
        router_id = self.router_info.router_id
        connection_id = FAKE_IPSEC_SITE_CONNECTION2_ID
        self.driver.ensure_process(router_id, self.vpnservice)
        self._execute.return_value = down_status
        self.driver.report_status(mock.Mock())
        process_status = self.driver.process_status_cache[router_id]
        ipsec_site_conn = process_status['ipsec_site_connections']
        self.assertEqual(constants.ACTIVE, process_status['status'])
        self.assertEqual(constants.DOWN,
                         ipsec_site_conn[connection_id]['status'])

    def _test_status_handling_for_active_connection(self, active_status):
        """Test status handling for active connection."""
        router_id = self.router_info.router_id
        connection_id = FAKE_IPSEC_SITE_CONNECTION2_ID
        self.driver.ensure_process(router_id, self.vpnservice)
        self._execute.return_value = active_status
        self.driver.report_status(mock.Mock())
        process_status = self.driver.process_status_cache[
            router_id]
        ipsec_site_conn = process_status['ipsec_site_connections']
        self.assertEqual(constants.ACTIVE, process_status['status'])
        self.assertEqual(constants.ACTIVE,
                         ipsec_site_conn[connection_id]['status'])

    def _test_status_handling_for_ike_v2_active_connection(self,
                                                           active_status):
        """Test status handling for active connection."""
        router_id = self.router_info.router_id
        connection_id = FAKE_IPSEC_SITE_CONNECTION2_ID
        ike_policy = {'ike_version': 'v2',
                      'encryption_algorithm': 'aes-128',
                      'auth_algorithm': 'sha1',
                      'pfs': 'group5',
                      'lifetime_value': 3600}
        vpn_service = FAKE_VPN_SERVICE
        for isc in vpn_service["ipsec_site_connections"]:
            isc['ikepolicy'] = ike_policy
        self.driver.ensure_process(router_id, vpn_service)
        self._execute.return_value = active_status
        self.driver.report_status(mock.Mock())
        process_status = self.driver.process_status_cache[
            router_id]
        ipsec_site_conn = process_status['ipsec_site_connections']
        self.assertEqual(constants.ACTIVE, process_status['status'])
        self.assertEqual(constants.ACTIVE,
                         ipsec_site_conn[connection_id]['status'])

    def _test_connection_names_handling_for_multiple_subnets(self,
                                                             active_status):
        """Test connection names handling for multiple subnets."""
        router_id = self.router_info.router_id
        process = self.driver.ensure_process(router_id, self.vpnservice)
        self._execute.return_value = active_status
        names = process.get_established_connections()
        self.assertEqual(2, len(names))

    def _test_status_handling_for_deleted_connection(self,
                                                     not_running_status):
        """Test status handling for deleted connection."""
        router_id = self.router_info.router_id
        self.driver.ensure_process(router_id, self.vpnservice)
        self._execute.return_value = not_running_status
        self.driver.report_status(mock.Mock())
        process_status = self.driver.process_status_cache[router_id]
        ipsec_site_conn = process_status['ipsec_site_connections']
        self.assertEqual(constants.DOWN, process_status['status'])
        self.assertFalse(ipsec_site_conn)

    def _test_parse_connection_status(self, not_running_status,
                                      active_status, down_status):
        """Test the status of ipsec-site-connection is parsed correctly."""
        router_id = self.router_info.router_id
        process = self.driver.ensure_process(router_id, self.vpnservice)
        self._execute.return_value = not_running_status
        self.assertFalse(process.active)
        # An empty return value to simulate that the process
        # does not have any status to report.
        self._execute.return_value = ''
        self.assertFalse(process.active)
        self._execute.return_value = active_status
        self.assertTrue(process.active)
        self._execute.return_value = down_status
        self.assertTrue(process.active)

    def test_get_namespace_for_router(self):
        namespace = self.driver.get_namespace(FAKE_ROUTER_ID)
        self.assertEqual('qrouter-' + FAKE_ROUTER_ID, namespace)

    def test_fail_getting_namespace_for_unknown_router(self):
        self.assertFalse(self.driver.get_namespace('bogus_id'))


class IPSecDeviceDVR(BaseIPsecDeviceDriver):

    def setUp(self, driver=openswan_ipsec.OpenSwanDriver,
              ipsec_process=openswan_ipsec.OpenSwanProcess):
        super().setUp(driver, ipsec_process)
        mock.patch.object(dvr_snat_ns.SnatNamespace, 'create').start()
        self._make_dvr_edge_router_info_for_test()

    def _make_dvr_edge_router_info_for_test(self):
        router_info = dvr_edge_router.DvrEdgeRouter(mock.sentinel.agent,
                                                    mock.sentinel.myhost,
                                                    FAKE_ROUTER_ID,
                                                    **self.ri_kwargs)
        router_info.router['distributed'] = True
        router_info.snat_namespace = dvr_snat_ns.SnatNamespace(
            router_info.router['id'],
            mock.sentinel.agent,
            self.driver,
            mock.ANY
        )
        router_info.snat_namespace.create()
        router_info.snat_iptables_manager = iptables_manager.IptablesManager(
            namespace='snat-' + FAKE_ROUTER_ID, use_ipv6=mock.ANY)
        router_info.snat_iptables_manager.ipv4['nat'] = self.iptables
        router_info.snat_iptables_manager.apply = self.apply_mock
        self.driver.routers[FAKE_ROUTER_ID] = router_info

    def test_sync_dvr(self):
        fake_vpn_service = FAKE_VPN_SERVICE
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
            fake_vpn_service]
        context = mock.Mock()
        self.driver._sync_vpn_processes = mock.Mock()
        self.driver._delete_vpn_processes = mock.Mock()
        self.driver._cleanup_stale_vpn_processes = mock.Mock()
        sync_router_ids = [fake_vpn_service['router_id']]
        with mock.patch.object(self.driver,
                               'get_process_status_cache') as process_status:
            self.driver.sync(context, [self.driver.routers[FAKE_ROUTER_ID]])
            self.driver._sync_vpn_processes.assert_called_once_with(
                [fake_vpn_service], sync_router_ids)
            self.driver._delete_vpn_processes.assert_called_once_with(
                sync_router_ids, sync_router_ids)
            self.driver._cleanup_stale_vpn_processes.assert_called_once_with(
                sync_router_ids)
            self.assertEqual(0, process_status.call_count)

    def test_get_namespace_for_dvr_edge_router(self):
        namespace = self.driver.get_namespace(FAKE_ROUTER_ID)
        self.assertEqual('snat-' + FAKE_ROUTER_ID, namespace)

    def test_ensure_nat_rules_with_dvr_edge_router(self):
        self.driver.ensure_nat_rules(FAKE_VPN_SERVICE)
        self.apply_mock.assert_called_once_with()


class TestOpenSwanConfigGeneration(BaseIPsecDeviceDriver):

    """Verify that configuration files are generated correctly.

    Besides the normal translation of some settings, when creating the config
    file, the generated file can also vary based on the following
    special conditions:

        - IPv6 versus IPv4
        - Multiple left subnets versus a single left subnet
        - IPSec policy using AH transform

    The tests will focus on these variations.
    """

    def setUp(self, driver=openswan_ipsec.OpenSwanDriver,
              ipsec_process=openswan_ipsec.OpenSwanProcess):
        super().setUp(driver, ipsec_process, vpnservice=FAKE_VPN_SERVICE)
        self.conf.register_opts(openswan_ipsec.openswan_opts, 'openswan')
        self.conf.set_override('state_path', '/tmp')
        self.ipsec_template = self.conf.openswan.ipsec_config_template
        self.process = ipsec_process(self.conf,
                                     'foo-process-id',
                                     self.vpnservice,
                                     mock.ANY)

    def build_ipsec_expected_config_for_test(self, info):
        """Modify OpenSwan ipsec expected config files for test variations."""
        auth_mode = info.get('ipsec_auth', AUTH_ESP)
        conn_details = OPENSWAN_CONNECTION_DETAILS % {
            'auth_mode': auth_mode,
            'dpd_action': 'hold',
            'dpd_delay': 30,
            'dpd_timeout': 120,
            'ike_lifetime': 3600,
            'life_time': 3600,
            'encapsulation_mode': 'tunnel'}
        virtual_privates = []
        # Convert local CIDRs into assignment strings. IF more than one,
        # pluralize the attribute name and enclose in brackets.
        cidrs = info.get('local_cidrs', [['10.0.0.0/24'], ['11.0.0.0/24']])
        local_cidrs = []
        for cidr in cidrs:
            if len(cidr) == 2:
                local_cidrs.append("s={ %s }" % ' '.join(cidr))
            else:
                local_cidrs.append("=%s" % cidr[0])
            for net in cidr:
                version = netaddr.IPNetwork(net).version
                virtual_privates.append('%v{}:{}'.format(version, net))
        # Convert peer CIDRs into space separated strings
        cidrs = info.get('peer_cidrs', [['20.0.0.0/24', '30.0.0.0/24'],
                                        ['40.0.0.0/24', '50.0.0.0/24']])
        for cidr in cidrs:
            for net in cidr:
                version = netaddr.IPNetwork(net).version
                virtual_privates.append('%v{}:{}'.format(version, net))
        peer_cidrs = [' '.join(cidr) for cidr in cidrs]
        local_ip = info.get('local', '60.0.0.4')
        local_id = info.get('local_id')
        leftid = local_ip
        if local_id:
            leftid = local_id
        version = info.get('local_ip_vers', 4)
        next_hop = IPV4_NEXT_HOP if version == 4 else IPV6_NEXT_HOP % local_ip
        peer_ips = info.get('peers', ['60.0.0.5', '60.0.0.6'])
        virtual_privates.sort()
        return EXPECTED_OPENSWAN_CONF % {
            'vpnservice_id': FAKE_VPNSERVICE_ID,
            'virtual_privates': ','.join(virtual_privates),
            'next_hop': next_hop,
            'local_cidrs1': local_cidrs[0], 'local_cidrs2': local_cidrs[1],
            'local_ver': version,
            'peer_cidrs1': peer_cidrs[0], 'peer_cidrs2': peer_cidrs[1],
            'left': local_ip,
            'leftid': leftid,
            'right1': peer_ips[0], 'right2': peer_ips[1],
            'conn1_id': FAKE_IPSEC_SITE_CONNECTION1_ID,
            'conn2_id': FAKE_IPSEC_SITE_CONNECTION2_ID,
            'conn_details': conn_details}

    def test_connections_with_esp_transform_protocol(self):
        """Test config file with IPSec policy using ESP."""
        self._test_ipsec_connection_config({})

    def test_connections_with_ah_transform_protocol(self):
        """Test config file with IPSec policy using ESP."""
        overrides = {'ipsec_auth': 'ah'}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        info = {'ipsec_auth': AUTH_AH}
        self._test_ipsec_connection_config(info)

    def test_connections_with_multiple_left_subnets(self):
        """Test multiple local subnets.

        The configure uses the 'leftsubnets' attribute, instead of the
        'leftsubnet' attribute.
        """

        overrides = {'local_cidrs': [['10.0.0.0/24', '11.0.0.0/24'],
                                     ['12.0.0.0/24', '13.0.0.0/24']]}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        self._test_ipsec_connection_config(overrides)

    def test_config_files_with_ipv6_addresses(self):
        """Test creating config files using IPv6 addressing."""
        overrides = {'local_cidrs': [['2002:0a00::/48'], ['2002:0b00::/48']],
                     'local_ip_vers': 6,
                     'peer_cidrs': [['2002:1400::/48', '2002:1e00::/48'],
                                    ['2002:2800::/48', '2002:3200::/48']],
                     'local': '2002:3c00:0004::',
                     'peers': ['2002:3c00:0005::', '2002:3c00:0006::'],
                     'local_id': '2002:3c00:0004::'}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        self._test_ipsec_connection_config(overrides)

    def test_config_files_with_ipv6_addresses_without_local_id(self):
        """Test creating config files using IPv6 addressing."""
        overrides = {'local_cidrs': [['2002:0a00::/48'], ['2002:0b00::/48']],
                     'local_ip_vers': 6,
                     'peer_cidrs': [['2002:1400::/48', '2002:1e00::/48'],
                                    ['2002:2800::/48', '2002:3200::/48']],
                     'local': '2002:3c00:0004::',
                     'peers': ['2002:3c00:0005::', '2002:3c00:0006::']}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        self._test_ipsec_connection_config(overrides)

    def test_secrets_config_file(self):
        expected = EXPECTED_IPSEC_OPENSWAN_SECRET_CONF
        actual = self.process._gen_config_content(
            self.conf.openswan.ipsec_secret_template, self.vpnservice)
        self.check_config_file(expected, actual)


class IPsecStrongswanConfigGeneration(BaseIPsecDeviceDriver):

    def setUp(self, driver=strongswan_ipsec.StrongSwanDriver,
              ipsec_process=strongswan_ipsec.StrongSwanProcess):
        super().setUp(
            driver, ipsec_process, vpnservice=FAKE_VPN_SERVICE)
        self.conf.register_opts(strongswan_ipsec.strongswan_opts,
                                'strongswan')
        self.conf.set_override('state_path', '/tmp')
        self.ipsec_template = self.conf.strongswan.ipsec_config_template
        self.process = strongswan_ipsec.StrongSwanProcess(self.conf,
                                                          'foo-process-id',
                                                          self.vpnservice,
                                                          mock.ANY)

    def build_ipsec_expected_config_for_test(self, info):
        cidrs = info.get('local_cidrs', [['10.0.0.0/24'], ['11.0.0.0/24']])
        local_cidrs = [','.join(cidr) for cidr in cidrs]
        cidrs = info.get('peer_cidrs', [['20.0.0.0/24', '30.0.0.0/24'],
                                        ['40.0.0.0/24', '50.0.0.0/24']])
        peer_cidrs = [','.join(cidr) for cidr in cidrs]
        local_ip = info.get('local', '60.0.0.4')
        local_id = info.get('local_id')
        leftid = local_ip
        if local_id:
            leftid = local_id
        peer_ips = info.get('peers', ['60.0.0.5', '60.0.0.6'])
        auth_mode = info.get('ipsec_auth', STRONGSWAN_AUTH_ESP)
        return EXPECTED_IPSEC_STRONGSWAN_CONF % {
            'vpnservice_id': FAKE_VPNSERVICE_ID,
            'local_cidrs1': local_cidrs[0], 'local_cidrs2': local_cidrs[1],
            'peer_cidrs1': peer_cidrs[0], 'peer_cidrs2': peer_cidrs[1],
            'left': local_ip,
            'leftid': leftid,
            'right1': peer_ips[0], 'right2': peer_ips[1],
            'dpd_action': 'hold',
            'dpd_delay': 30,
            'dpd_timeout': 120,
            'ike_encryption_algorithm': 'aes128',
            'ike_auth_algorithm': 'sha1',
            'ike_pfs': 'modp1536',
            'ike_lifetime': 3600,
            'life_time': 3600,
            'auth_mode': auth_mode,
            'encapsulation_mode': 'tunnel',
            'conn1_id': FAKE_IPSEC_SITE_CONNECTION1_ID,
            'conn2_id': FAKE_IPSEC_SITE_CONNECTION2_ID}

    def test_ipsec_config_file_with_esp(self):
        self._test_ipsec_connection_config({})

    def test_ipsec_config_file_with_ah(self):
        overrides = {'ipsec_auth': 'ah'}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        info = {'ipsec_auth': STRONGSWAN_AUTH_AH}
        self._test_ipsec_connection_config(info)

    def test_ipsec_config_file_for_v6(self):
        overrides = {'local_cidrs': [['2002:0a00::/48'], ['2002:0b00::/48']],
                     'peer_cidrs': [['2002:1400::/48', '2002:1e00::/48'],
                                    ['2002:2800::/48', '2002:3200::/48']],
                     'local': '2002:3c00:0004::',
                     'peers': ['2002:3c00:0005::', '2002:3c00:0006::'],
                     'local_id': '2002:3c00:0004::'}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        self._test_ipsec_connection_config(overrides)

    def test_ipsec_config_file_for_v6_without_local_id(self):
        overrides = {'local_cidrs': [['2002:0a00::/48'], ['2002:0b00::/48']],
                     'peer_cidrs': [['2002:1400::/48', '2002:1e00::/48'],
                                    ['2002:2800::/48', '2002:3200::/48']],
                     'local': '2002:3c00:0004::',
                     'peers': ['2002:3c00:0005::', '2002:3c00:0006::']}
        self.modify_config_for_test(overrides)
        self.process.update_vpnservice(self.vpnservice)
        self._test_ipsec_connection_config(overrides)

    def test_strongswan_default_config_file(self):
        expected = EXPECTED_STRONGSWAN_DEFAULT_CONF
        actual = self.process._gen_config_content(
            self.conf.strongswan.strongswan_config_template, self.vpnservice)
        self.check_config_file(expected, actual)

    def test_secrets_config_file(self):
        expected = EXPECTED_IPSEC_STRONGSWAN_SECRET_CONF
        actual = self.process._gen_config_content(
            self.conf.strongswan.ipsec_secret_template, self.vpnservice)
        self.check_config_file(expected, actual)


class TestLibreSwanConfigGeneration(TestOpenSwanConfigGeneration):
    def setUp(self, driver=libreswan_ipsec.LibreSwanDriver,
              ipsec_process=libreswan_ipsec.LibreSwanProcess):
        super().setUp(driver=driver, ipsec_process=ipsec_process)

    def build_ipsec_expected_config_for_test(self, info):
        expected = super().build_ipsec_expected_config_for_test(info)
        return expected.replace('    nat_traversal=yes\n', '')


class TestOpenSwanProcess(IPSecDeviceLegacy):

    _test_timeout = 1
    _test_backoff = 2
    _test_retries = 5

    def setUp(self, driver=openswan_ipsec.OpenSwanDriver,
              ipsec_process=openswan_ipsec.OpenSwanProcess):
        super().setUp(driver, ipsec_process)
        self.conf.register_opts(openswan_ipsec.openswan_opts,
                                'openswan')
        self.conf.set_override('state_path', '/tmp')
        cfg.CONF.register_opts(openswan_ipsec.pluto_opts,
                               'pluto')
        cfg.CONF.set_override('shutdown_check_timeout', self._test_timeout,
                              group='pluto')
        cfg.CONF.set_override('shutdown_check_back_off', self._test_backoff,
                              group='pluto')
        cfg.CONF.set_override('shutdown_check_retries', self._test_retries,
                              group='pluto')
        self.addCleanup(cfg.CONF.reset)

        self.os_remove = mock.patch('os.remove').start()

        self.process = openswan_ipsec.OpenSwanProcess(self.conf,
                                                      'foo-process-id',
                                                      self.vpnservice,
                                                      mock.ANY)

    def test__resolve_fqdn(self):
        with mock.patch.object(socket, 'getaddrinfo') as mock_getaddr_info:
            mock_getaddr_info.return_value = [(2, 1, 6, '',
                                              ('172.168.1.2', 0))]
            resolved_ip_addr = self.process._resolve_fqdn('fqdn.foo.addr')
            self.assertEqual('172.168.1.2', resolved_ip_addr)

    def _test_get_nexthop_helper(self, address, _resolve_fqdn_side_effect,
                                 expected_ip_cmd, expected_nexthop):
        with mock.patch.object(self.process,
                               '_resolve_fqdn') as fake_resolve_fqdn:
            fake_resolve_fqdn.side_effect = _resolve_fqdn_side_effect

            returned_next_hop = self.process._get_nexthop(address,
                                                          'fake-conn-id')
            _resolve_fqdn_expected_call_count = (
                1 if _resolve_fqdn_side_effect else 0)

            self.assertEqual(_resolve_fqdn_expected_call_count,
                             fake_resolve_fqdn.call_count)
            self._execute.assert_called_once_with(expected_ip_cmd)
            self.assertEqual(expected_nexthop, returned_next_hop)

    def test__get_nexthop_peer_addr_is_ipaddr(self):
        gw_addr = '10.0.0.1'
        self._execute.return_value = '172.168.1.2 via %s' % gw_addr
        peer_address = '172.168.1.2'
        expected_ip_cmd = ['ip', 'route', 'get', peer_address]
        self._test_get_nexthop_helper(peer_address, None,
                                      expected_ip_cmd, gw_addr)

    def test__get_nexthop_peer_addr_is_valid_fqdn(self):
        peer_address = 'foo.peer.addr'
        expected_ip_cmd = ['ip', 'route', 'get', '172.168.1.2']
        gw_addr = '10.0.0.1'
        self._execute.return_value = '172.168.1.2 via %s' % gw_addr

        def _fake_resolve_fqdn(address):
            return '172.168.1.2'

        self._test_get_nexthop_helper(peer_address, _fake_resolve_fqdn,
                                      expected_ip_cmd, gw_addr)

    def test__get_nexthop_gw_not_present(self):
        peer_address = '172.168.1.2'
        expected_ip_cmd = ['ip', 'route', 'get', '172.168.1.2']
        self._execute.return_value = ' '

        self._test_get_nexthop_helper(peer_address, None,
                                      expected_ip_cmd, peer_address)

    def test__get_nexthop_fqdn_peer_addr_is_not_resolved(self):
        self.process.connection_status = {}
        expected_connection_status_dict = (
            {'fake-conn-id': {'status': constants.ERROR,
                              'updated_pending_status': True}})

        self.assertRaises(vpn_exception.VPNPeerAddressNotResolved,
                          self.process._get_nexthop, 'foo.peer.addr.',
                          'fake-conn-id')
        self.assertEqual(expected_connection_status_dict,
                         self.process.connection_status)

        self.process.connection_status = (
            {'fake-conn-id': {'status': constants.PENDING_CREATE,
                              'updated_pending_status': False}})

        self.assertRaises(vpn_exception.VPNPeerAddressNotResolved,
                          self.process._get_nexthop, 'foo.peer.addr.',
                          'fake-conn-id')
        self.assertEqual(expected_connection_status_dict,
                         self.process.connection_status)

    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._get_nexthop',
                return_value='172.168.1.2')
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._cleanup_control_files')
    def test_no_cleanups(self, cleanup_mock, hop_mock):
        # Not an "awesome test" but more of a check box item. Basically,
        # what happens if we didn't need to clean up any files.
        with mock.patch.object(self.process,
                               '_process_running',
                               return_value=True) as query_mock:
            self.process.start()
            self.assertEqual(1, query_mock.call_count)

            # This is really what is being tested here. If process is
            # running, we shouldn't attempt a cleanup.
            self.assertFalse(cleanup_mock.called)

    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._get_nexthop',
                return_value='172.168.1.2')
    @mock.patch('os.path.exists', return_value=True)
    def test_cleanup_files(self, exists_mock, hop_mock):
        # Tests the 'bones' of things really and kind of check-box-item-bogus
        # test - this really needs exercising through a higher level test.
        with mock.patch.object(self.process,
                               '_process_running',
                               return_value=False) as query_mock:
            fake_path = '/fake/path/run'
            self.process.pid_path = fake_path
            self.process.pid_file = '%s.pid' % fake_path
            self.process.start()
            self.assertEqual(1, query_mock.call_count)
            self.assertEqual(2, self.os_remove.call_count)
            self.os_remove.assert_has_calls([mock.call('%s.pid' % fake_path),
                                             mock.call('%s.ctl' % fake_path)])

    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._get_nexthop',
                return_value='172.168.1.2')
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._process_running',
                return_value=False)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._cleanup_control_files')
    @mock.patch('eventlet.sleep')
    def test_restart_process_not_running(self, sleep_mock, cleanup_mock,
                                         query_mock, hop_mock):
        self.process.restart()

        # Really what is being tested - retry configuration exists and that
        # we do the right things when process check is false.
        self.assertTrue(query_mock.called)
        self.assertTrue(cleanup_mock.called)
        self.assertFalse(sleep_mock.called)

    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._get_nexthop',
                return_value='172.168.1.2')
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._process_running',
                return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._cleanup_control_files')
    @mock.patch('eventlet.sleep')
    def test_restart_process_doesnt_stop(self, sleep_mock, cleanup_mock,
                                         query_mock, hop_mock):
        self.process.restart()

        # Really what is being tested - retry configuration exists and that
        # we do the right things when process check is True.
        self.assertEqual(self._test_retries + 1, query_mock.call_count)
        self.assertFalse(cleanup_mock.called)
        self.assertEqual(self._test_retries, sleep_mock.call_count)
        calls = [mock.call(1), mock.call(2), mock.call(4),
                 mock.call(8), mock.call(16)]
        sleep_mock.assert_has_calls(calls)

    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._get_nexthop',
                return_value='172.168.1.2')
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._process_running',
                side_effect=[True, True, False, False])
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.'
                'ipsec.OpenSwanProcess._cleanup_control_files')
    @mock.patch('eventlet.sleep')
    def test_restart_process_retry_until_stop(self, sleep_mock, cleanup_mock,
                                              query_mock, hop_mock):
        self.process.restart()

        # Really what is being tested - retry configuration exists and that
        # we do the right things when process check is True a few times and
        # then returns False.
        self.assertEqual(4, query_mock.call_count)
        self.assertTrue(cleanup_mock.called)
        self.assertEqual(2, sleep_mock.call_count)

    def test_process_running_no_pid(self):
        with mock.patch('os.path.exists', return_value=False):
            self.assertFalse(
                self.process._process_running())

    # open() is used elsewhere, so we need to inject a mocked open into the
    # module to be tested.
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.open',
                create=True,
                side_effect=IOError)
    def test_process_running_open_failure(self, mock_open, mock_exists):
        self.assertFalse(self.process._process_running())
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_open.called)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.open',
                create=True,
                side_effect=[io.StringIO('invalid'),
                             IOError])
    def test_process_running_bogus_pid(self, mock_open, mock_exists):
        with mock.patch.object(openswan_ipsec.LOG, 'error'):
            self.assertFalse(self.process._process_running())
            self.assertTrue(mock_exists.called)
            self.assertEqual(2, mock_open.call_count)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.open',
                create=True,
                side_effect=[io.StringIO('134'), io.StringIO('')])
    def test_process_running_no_cmdline(self, mock_open, mock_exists):
        with mock.patch.object(openswan_ipsec.LOG, 'error') as log_mock:
            self.assertFalse(self.process._process_running())
            self.assertFalse(log_mock.called)
            self.assertEqual(2, mock_open.call_count)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.open',
                create=True,
                side_effect=[io.StringIO('134'), io.StringIO('ps ax')])
    def test_process_running_cmdline_mismatch(self, mock_open, mock_exists):
        with mock.patch.object(openswan_ipsec.LOG, 'error') as log_mock:
            self.assertFalse(self.process._process_running())
            self.assertFalse(log_mock.called)
            self.assertEqual(2, mock_open.call_count)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.open',
                create=True,
                side_effect=[io.StringIO('134'),
                             io.StringIO('/usr/libexec/ipsec/pluto -ctlbase'
                                         '/some/foo/path')])
    def test_process_running_cmdline_match(self, mock_open, mock_exists):
        self.process.pid_path = '/some/foo/path'
        with mock.patch.object(openswan_ipsec.LOG, 'error') as log_mock:
            self.assertTrue(self.process._process_running())
            self.assertTrue(log_mock.called)

    def test_status_handling_for_downed_connection(self):
        """Test status handling for downed connection."""
        self._test_status_handling_for_downed_connection(PLUTO_DOWN_STATUS)

    def test_status_handling_for_connection_with_no_ipsec_sa(self):
        """Test status handling for downed connection."""
        self._test_status_handling_for_downed_connection(
            PLUTO_ACTIVE_NO_IPSEC_SA_STATUS)

    def test_status_handling_for_active_connection(self):
        """Test status handling for active connection."""
        self._test_status_handling_for_active_connection(PLUTO_ACTIVE_STATUS)

    def test_status_handling_for_ike_v2_active_connection(self):
        """Test status handling for active connection."""
        self._test_status_handling_for_ike_v2_active_connection(
            PLUTO_ACTIVE_STATUS_IKEV2)

    def test_status_handling_for_deleted_connection(self):
        """Test status handling for deleted connection."""
        self._test_status_handling_for_deleted_connection(NOT_RUNNING_STATUS)

    def test_connection_names_handling_for_multiple_subnets(self):
        """Test connection names handling for multiple subnets."""
        self._test_connection_names_handling_for_multiple_subnets(
            PLUTO_MULTIPLE_SUBNETS_ESTABLISHED_STATUS)

    def test_parse_connection_status(self):
        """Test the status of ipsec-site-connection parsed correctly."""
        self._test_parse_connection_status(NOT_RUNNING_STATUS,
                                           PLUTO_ACTIVE_STATUS,
                                           PLUTO_DOWN_STATUS)


class TestLibreSwanProcess(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.vpnservice = copy.deepcopy(FAKE_VPN_SERVICE)

        self.ipsec_process = libreswan_ipsec.LibreSwanProcess(cfg.CONF,
                                                              'foo-process-id',
                                                              self.vpnservice,
                                                              mock.ANY)

    @mock.patch('os.path.exists', return_value=True)
    def test_ensure_configs_on_restart(self, exists_mock):
        openswan_ipsec.OpenSwanProcess.ensure_configs = mock.Mock()
        with mock.patch.object(
            self.ipsec_process, '_execute'
        ) as fake_execute, mock.patch.object(
            self.ipsec_process, '_ipsec_execute'
        ) as fake_ipsec_execute, mock.patch.object(
            self.ipsec_process, '_ensure_needed_files'
        ) as fake_ensure_needed_files:
            self.ipsec_process.ensure_configs()

            expected = [mock.call(['rm', '-f',
                                   self.ipsec_process._get_config_filename(
                                       'ipsec.secrets')]),
                        mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root',
                                   self.ipsec_process._get_config_filename(
                                       'ipsec.secrets')]),
                        mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root', self.ipsec_process.log_dir])]
            fake_execute.assert_has_calls(expected)
            self.assertEqual(3, fake_execute.call_count)

            expected = [mock.call(['_stackmanager', 'start']),
                        mock.call(['checknss'])]
            fake_ipsec_execute.assert_has_calls(expected)
            self.assertEqual(2, fake_ipsec_execute.call_count)

            self.assertTrue(fake_ensure_needed_files.called)
            self.assertTrue(exists_mock.called)

    @mock.patch('os.path.exists', return_value=False)
    def test_ensure_configs(self, exists_mock):
        openswan_ipsec.OpenSwanProcess.ensure_configs = mock.Mock()
        with mock.patch.object(
            self.ipsec_process, '_execute'
        ) as fake_execute, mock.patch.object(
            self.ipsec_process, '_ipsec_execute'
        ) as fake_ipsec_execute, mock.patch.object(
            self.ipsec_process, '_ensure_needed_files'
        ) as fake_ensure_needed_files:
            self.ipsec_process.ensure_configs()

            expected = [mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root',
                                   self.ipsec_process._get_config_filename(
                                       'ipsec.secrets')]),
                        mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root', self.ipsec_process.log_dir])]
            fake_execute.assert_has_calls(expected)
            self.assertEqual(2, fake_execute.call_count)

            expected = [mock.call(['_stackmanager', 'start']),
                        mock.call(['checknss'])]
            fake_ipsec_execute.assert_has_calls(expected)
            self.assertEqual(2, fake_ipsec_execute.call_count)

            self.assertTrue(fake_ensure_needed_files.called)
            self.assertTrue(exists_mock.called)

        exists_mock.reset_mock()

        with mock.patch.object(
            self.ipsec_process, '_execute'
        ) as fake_execute, mock.patch.object(
            self.ipsec_process, '_ipsec_execute'
        ) as fake_ipsec_execute, mock.patch.object(
            self.ipsec_process, '_ensure_needed_files'
        ) as fake_ensure_needed_files:
            fake_ipsec_execute.side_effect = [None, RuntimeError, None]
            self.ipsec_process.ensure_configs()

            expected = [mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root',
                                   self.ipsec_process._get_config_filename(
                                       'ipsec.secrets')]),
                        mock.call(['chown', '--from=%s' % os.getuid(),
                                   'root:root', self.ipsec_process.log_dir])]
            fake_execute.assert_has_calls(expected)
            self.assertEqual(2, fake_execute.call_count)

            expected = [mock.call(['_stackmanager', 'start']),
                        mock.call(['checknss']),
                        mock.call(['initnss'])]
            self.assertEqual(3, fake_ipsec_execute.call_count)
            fake_ipsec_execute.assert_has_calls(expected)

            self.assertTrue(fake_ensure_needed_files.called)
            self.assertTrue(exists_mock.called)


class IPsecStrongswanDeviceDriverLegacy(IPSecDeviceLegacy):

    def setUp(self, driver=strongswan_ipsec.StrongSwanDriver,
              ipsec_process=strongswan_ipsec.StrongSwanProcess):
        super().setUp(driver, ipsec_process)
        self.conf.register_opts(strongswan_ipsec.strongswan_opts,
                                'strongswan')
        self.conf.set_override('state_path', '/tmp')
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
            self.vpnservice]

    def test_status_handling_for_downed_connection(self):
        """Test status handling for downed connection."""
        self._test_status_handling_for_downed_connection(CHARON_DOWN_STATUS)

    def test_status_handling_for_active_connection(self):
        """Test status handling for active connection."""
        self._test_status_handling_for_active_connection(CHARON_ACTIVE_STATUS)

    def test_status_handling_for_deleted_connection(self):
        """Test status handling for deleted connection."""
        self._test_status_handling_for_deleted_connection(NOT_RUNNING_STATUS)

    def test_parse_connection_status(self):
        """Test the status of ipsec-site-connection parsed correctly."""
        self._test_parse_connection_status(NOT_RUNNING_STATUS,
                                           CHARON_ACTIVE_STATUS,
                                           CHARON_DOWN_STATUS)


class IPsecStrongswanDeviceDriverDVR(IPSecDeviceDVR):
    def setUp(self, driver=strongswan_ipsec.StrongSwanDriver,
              ipsec_process=strongswan_ipsec.StrongSwanProcess):
        super().setUp(driver, ipsec_process)



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.db import servicetype_db as st_db
from neutron_lib import context as n_ctx
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils

from neutron_vpnaas.services.vpn import plugin as vpn_plugin
from neutron_vpnaas.services.vpn.service_drivers import ipsec as ipsec_driver
from neutron_vpnaas.services.vpn.service_drivers import ipsec_validator
from neutron_vpnaas.tests import base


_uuid = uuidutils.generate_uuid

FAKE_SERVICE_ID = _uuid()
FAKE_VPN_CONNECTION = {
    'vpnservice_id': FAKE_SERVICE_ID
}
FAKE_ROUTER_ID = _uuid()
FAKE_VPN_SERVICE = {
    'router_id': FAKE_ROUTER_ID
}
FAKE_HOST = 'fake_host'
FAKE_CONN_ID = _uuid()

IPSEC_SERVICE_DRIVER = ('neutron_vpnaas.services.vpn.service_drivers.'
                        'ipsec.IPsecVPNDriver')


class FakeSqlQueryObject(dict):
    """To fake SqlAlchemy query object and access keys as attributes."""

    def __init__(self, **entries):
        self.__dict__.update(entries)
        super().__init__(**entries)


class TestValidatorSelection(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        vpnaas_provider = [{
            'service_type': constants.VPN,
            'name': 'vpnaas',
            'driver': IPSEC_SERVICE_DRIVER,
            'default': True
        }]
        # override the default service provider
        self.service_providers = (
            mock.patch.object(st_db.ServiceTypeManager,
                              'get_service_providers').start())
        self.service_providers.return_value = vpnaas_provider
        mock.patch('neutron_lib.rpc.Connection').start()
        stm = st_db.ServiceTypeManager()
        stm.get_provider_names_by_resource_ids = mock.Mock(
            return_value={})
        mock.patch('neutron.db.servicetype_db.ServiceTypeManager.get_instance',
                   return_value=stm).start()
        mock.patch('neutron_vpnaas.db.vpn.vpn_db.VPNPluginDb.get_vpnservices',
                   return_value=[]).start()
        self.vpn_plugin = vpn_plugin.VPNDriverPlugin()

    def test_reference_driver_used(self):
        default_provider = self.vpn_plugin.default_provider
        default_driver = self.vpn_plugin.drivers[default_provider]
        self.assertIsInstance(default_driver.validator,
                              ipsec_validator.IpsecVpnValidator)


class TestIPsecDriver(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        mock.patch('neutron_lib.rpc.Connection').start()

        l3_agent = mock.Mock()
        l3_agent.host = FAKE_HOST
        plugin = mock.Mock()
        plugin.get_l3_agents_hosting_routers.return_value = [l3_agent]
        directory.add_plugin(constants.CORE, plugin)
        directory.add_plugin(constants.L3, plugin)
        self.svc_plugin = mock.Mock()
        self.svc_plugin.get_l3_agents_hosting_routers.return_value = [l3_agent]
        self._fake_vpn_router_id = _uuid()
        self.svc_plugin.get_vpnservice_router_id.return_value = \
            self._fake_vpn_router_id
        self.driver = ipsec_driver.IPsecVPNDriver(self.svc_plugin)
        self.validator = ipsec_validator.IpsecVpnValidator(self.driver)
        self.context = n_ctx.get_admin_context()

    def _test_update(self, func, args, additional_info=None):
        ctxt = n_ctx.Context('', 'somebody')
        with mock.patch.object(self.driver.agent_rpc.client, 'cast'
                               ) as rpc_mock, \
                mock.patch.object(self.driver.agent_rpc.client, 'prepare'
                                  ) as prepare_mock:
            prepare_mock.return_value = self.driver.agent_rpc.client
            func(ctxt, *args)

        prepare_args = {'server': 'fake_host', 'version': '1.0'}
        prepare_mock.assert_called_once_with(**prepare_args)

        rpc_mock.assert_called_once_with(ctxt, 'vpnservice_updated',
                                         **additional_info)

    def test_create_ipsec_site_connection(self):
        self._test_update(self.driver.create_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION],
                          {'router': {'id': self._fake_vpn_router_id}})

    def test_update_ipsec_site_connection(self):
        self._test_update(self.driver.update_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION, FAKE_VPN_CONNECTION],
                          {'router': {'id': self._fake_vpn_router_id}})

    def test_delete_ipsec_site_connection(self):
        self._test_update(self.driver.delete_ipsec_site_connection,
                          [FAKE_VPN_CONNECTION],
                          {'router': {'id': self._fake_vpn_router_id}})

    def test_update_vpnservice(self):
        self._test_update(self.driver.update_vpnservice,
                          [FAKE_VPN_SERVICE, FAKE_VPN_SERVICE],
                          {'router': {'id': FAKE_VPN_SERVICE['router_id']}})

    def test_delete_vpnservice(self):
        self._test_update(self.driver.delete_vpnservice,
                          [FAKE_VPN_SERVICE],
                          {'router': {'id': FAKE_VPN_SERVICE['router_id']}})

    def prepare_dummy_query_objects(self, info):
        """Create fake query objects to test dict creation for sync oper."""
        external_ip = '10.0.0.99'
        peer_address = '10.0.0.2'
        peer_endpoints = info.get('peer_endpoints', [])
        local_endpoints = info.get('local_endpoints', [])
        peer_cidrs = info.get('peer_cidrs', ['40.4.0.0/24', '50.5.0.0/24'])
        peer_id = info.get('peer_id', '30.30.0.0')
        local_id = info.get('local_id', '')

        fake_ikepolicy = FakeSqlQueryObject(id='foo-ike', name='ike-name')
        fake_ipsecpolicy = FakeSqlQueryObject(id='foo-ipsec')

        fake_peer_cidrs_list = [
            FakeSqlQueryObject(cidr=cidr, ipsec_site_connection_id='conn-id')
            for cidr in peer_cidrs]

        peer_epg_id = 'peer-epg-id' if peer_endpoints else None
        local_epg_id = 'local-epg-id' if local_endpoints else None

        fake_ipsec_conn = FakeSqlQueryObject(id='conn-id',
                                             peer_id=peer_id,
                                             peer_address=peer_address,
                                             local_id=local_id,
                                             ikepolicy=fake_ikepolicy,
                                             ipsecpolicy=fake_ipsecpolicy,
                                             peer_ep_group_id=peer_epg_id,
                                             local_ep_group_id=local_epg_id,
                                             peer_cidrs=fake_peer_cidrs_list)

        if peer_endpoints:
            fake_peer_ep_group = FakeSqlQueryObject(id=peer_epg_id)
            fake_peer_ep_group.endpoints = [
                FakeSqlQueryObject(endpoint=ep,
                                   endpoint_group_id=peer_epg_id)
                for ep in peer_endpoints]
            fake_ipsec_conn.peer_ep_group = fake_peer_ep_group

        if local_endpoints:
            fake_local_ep_group = FakeSqlQueryObject(id=local_epg_id)
            fake_local_ep_group.endpoints = [
                FakeSqlQueryObject(endpoint=ep,
                                   endpoint_group_id=local_epg_id)
                for ep in local_endpoints]
            fake_ipsec_conn.local_ep_group = fake_local_ep_group
            subnet_id = None
        else:
            subnet_id = 'foo-subnet-id'

        fake_gw_port = {'fixed_ips': [{'ip_address': external_ip}]}
        fake_router = FakeSqlQueryObject(gw_port=fake_gw_port)
        fake_vpnservice = FakeSqlQueryObject(id='foo-vpn-id', name='foo-vpn',
                                             description='foo-vpn-service',
                                             admin_state_up=True,
                                             status='active',
                                             external_v4_ip=external_ip,
                                             external_v6_ip=None,
                                             subnet_id=subnet_id,
                                             router_id='foo-router-id',
                                             project_id='foo-project-id')
        if local_endpoints:
            fake_vpnservice.subnet = None
        else:
            fake_subnet = FakeSqlQueryObject(id=subnet_id,
                                             name='foo-subnet',
                                             cidr='9.0.0.0/16',
                                             network_id='foo-net-id')
            fake_vpnservice.subnet = fake_subnet
        fake_vpnservice.router = fake_router
        fake_vpnservice.ipsec_site_connections = [fake_ipsec_conn]
        return fake_vpnservice

    def build_expected_dict(self, info):
        """Create the expected dict used in sync operations.

        The default is to use non-endpoint groups, where the peer CIDRs come
        from the peer_cidrs arguments, the local CIDRs come from the (sole)
        subnet CIDR, and there is subnet info. Tests will customize the peer
        ID and peer CIDRs.
        """

        external_ip = '10.0.0.99'
        peer_id = info.get('peer_id', '30.30.0.0')
        peer_cidrs = info.get('peer_cidrs', ['40.4.0.0/24', '50.5.0.0/24'])
        local_id = info.get('local_id', '')

        return {'name': 'foo-vpn',
                'id': 'foo-vpn-id',
                'description': 'foo-vpn-service',
                'admin_state_up': True,
                'status': 'active',
                'external_v4_ip': external_ip,
                'external_v6_ip': None,
                'router_id': 'foo-router-id',
                'subnet': {'cidr': '9.0.0.0/16',
                           'id': 'foo-subnet-id',
                           'name': 'foo-subnet',
                           'network_id': 'foo-net-id'},
                'subnet_id': 'foo-subnet-id',
                'external_ip': external_ip,
                'project_id': 'foo-project-id',
                'tenant_id': 'foo-project-id',
                'ipsec_site_connections': [
                    {'id': 'conn-id',
                     'peer_id': peer_id,
                     'external_ip': external_ip,
                     'peer_address': '10.0.0.2',
                     'local_id': local_id,
                     'ikepolicy': {'id': 'foo-ike',
                                   'name': 'ike-name'},
                     'ipsecpolicy': {'id': 'foo-ipsec'},
                     'peer_ep_group_id': None,
                     'local_ep_group_id': None,
                     'peer_cidrs': peer_cidrs,
                     'local_cidrs': ['9.0.0.0/16'],
                     'local_ip_vers': 4}
                ]}

    def build_expected_dict_for_endpoints(self, info):
        """Create the expected dict used in sync operations for endpoints.

        The local and peer CIDRs come from the endpoint groups (with the
        local CIDR translated from the corresponding subnets specified).
        Tests will customize CIDRs, and the subnet, which is needed for
        backward compatibility with agents, during rolling upgrades.
        """

        external_ip = '10.0.0.99'
        peer_id = '30.30.0.0'
        local_id = info.get('local_id', '')
        return {'name': 'foo-vpn',
                'id': 'foo-vpn-id',
                'description': 'foo-vpn-service',
                'admin_state_up': True,
                'status': 'active',
                'external_v4_ip': external_ip,
                'external_v6_ip': None,
                'router_id': 'foo-router-id',
                'subnet': None,
                'subnet_id': None,
                'external_ip': external_ip,
                'project_id': 'foo-project-id',
                'tenant_id': 'foo-project-id',
                'ipsec_site_connections': [
                    {'id': 'conn-id',
                     'peer_id': peer_id,
                     'external_ip': external_ip,
                     'peer_address': '10.0.0.2',
                     'local_id': local_id,
                     'ikepolicy': {'id': 'foo-ike',
                                   'name': 'ike-name'},
                     'ipsecpolicy': {'id': 'foo-ipsec'},
                     'peer_ep_group_id': 'peer-epg-id',
                     'local_ep_group_id': 'local-epg-id',
                     'peer_cidrs': info['peers'],
                     'local_cidrs': info['locals'],
                     'local_ip_vers': info['vers']}
                ]}

    def test_make_vpnservice_dict_peer_id_is_ipaddr(self):
        """Peer ID as IP should be copied as-is, when creating dict."""
        subnet_cidr_map = {}
        peer_id_as_ip = {'peer_id': '10.0.0.2'}
        fake_service = self.prepare_dummy_query_objects(peer_id_as_ip)
        expected_dict = self.build_expected_dict(peer_id_as_ip)
        actual_dict = self.driver.make_vpnservice_dict(fake_service,
                                                       subnet_cidr_map)
        self.assertEqual(expected_dict, actual_dict)

        # make sure that ipsec_site_conn peer_id is not updated by
        # _make_vpnservice_dict (bug #1423244)
        self.assertEqual(peer_id_as_ip['peer_id'],
                         fake_service.ipsec_site_connections[0].peer_id)

    def test_make_vpnservice_dict_peer_id_is_string(self):
        """Peer ID as string should have '@' prepended, when creating dict."""
        subnet_cidr_map = {}
        peer_id_as_name = {'peer_id': 'foo.peer.id'}
        fake_service = self.prepare_dummy_query_objects(peer_id_as_name)
        expected_peer_id = {'peer_id': '@foo.peer.id'}
        expected_dict = self.build_expected_dict(expected_peer_id)
        actual_dict = self.driver.make_vpnservice_dict(fake_service,
                                                       subnet_cidr_map)
        self.assertEqual(expected_dict, actual_dict)

        # make sure that ipsec_site_conn peer_id is not updated by
        # _make_vpnservice_dict (bug #1423244)
        self.assertEqual(peer_id_as_name['peer_id'],
                         fake_service.ipsec_site_connections[0].peer_id)

    def test_make_vpnservice_dict_peer_cidrs_from_peer_cidr_table(self):
        """Peer CIDRs list populated from peer_cidr table.

        User provides peer CIDRs as parameters to IPSec site-to-site
        connection API, and they are stored in the peercidrs table.
        """
        subnet_cidr_map = {}
        peer_cidrs = {'peer_cidrs': ['80.0.0.0/24', '90.0.0.0/24']}
        fake_service = self.prepare_dummy_query_objects(peer_cidrs)
        expected_dict = self.build_expected_dict(peer_cidrs)
        actual_dict = self.driver.make_vpnservice_dict(fake_service,
                                                       subnet_cidr_map)
        self.assertEqual(expected_dict, actual_dict)

    def test_make_vpnservice_dict_cidrs_from_endpoints(self):
        """CIDRs list populated from local and peer endpoints.

        User provides peer and local endpoint group IDs in the IPSec
        site-to-site connection API. The endpoint groups contains peer
        CIDRs and local subnets (which will be mapped to CIDRs).
        """
        # Cannot have peer CIDRs specified, when using endpoint group
        subnet_cidr_map = {'local-sn1': '5.0.0.0/16',
                           'local-sn2': '5.1.0.0/16'}
        endpoint_groups = {'peer_cidrs': [],
                           'peer_endpoints': ['80.0.0.0/24', '90.0.0.0/24'],
                           'local_endpoints': ['local-sn1', 'local-sn2']}

        expected_cidrs = {'peers': ['80.0.0.0/24', '90.0.0.0/24'],
                          'locals': ['5.0.0.0/16', '5.1.0.0/16'],
                          'vers': 4}
        fake_service = self.prepare_dummy_query_objects(endpoint_groups)
        expected_dict = self.build_expected_dict_for_endpoints(expected_cidrs)
        expected_dict['subnet'] = {'cidr': '5.0.0.0/16'}
        actual_dict = self.driver.make_vpnservice_dict(fake_service,
                                                       subnet_cidr_map)
        self.assertEqual(expected_dict, actual_dict)

    def test_make_vpnservice_dict_v6_cidrs_from_endpoints(self):
        """IPv6 CIDRs list populated from local and peer endpoints."""
        # Cannot have peer CIDRs specified, when using endpoint group
        subnet_cidr_map = {'local-sn1': '2002:0a00:0000::/48',
                           'local-sn2': '2002:1400:0000::/48'}
        endpoint_groups = {'peer_cidrs': [],
                           'peer_endpoints': ['2002:5000:0000::/48',
                                              '2002:5a00:0000::/48'],
                           'local_endpoints': ['local-sn1', 'local-sn2']}

        expected_cidrs = {'peers': ['2002:5000:0000::/48',
                                    '2002:5a00:0000::/48'],
                          'locals': ['2002:0a00:0000::/48',
                                     '2002:1400:0000::/48'],
                          'vers': 6}
        fake_service = self.prepare_dummy_query_objects(endpoint_groups)
        expected_dict = self.build_expected_dict_for_endpoints(expected_cidrs)
        expected_dict['subnet'] = {'cidr': '2002:0a00:0000::/48'}
        actual_dict = self.driver.make_vpnservice_dict(fake_service,
                                                       subnet_cidr_map)
        self.assertEqual(expected_dict, actual_dict)

    def test_get_external_ip_based_on_ipv4_peer(self):
        vpnservice = mock.Mock()
        vpnservice.external_v4_ip = '10.0.0.99'
        vpnservice.external_v6_ip = '2001::1'
        ipsec_sitecon = {'id': FAKE_CONN_ID, 'peer_address': '10.0.0.9'}
        ip_to_use = self.driver.get_external_ip_based_on_peer(vpnservice,
                                                              ipsec_sitecon)
        self.assertEqual('10.0.0.99', ip_to_use)

    def test_get_external_ip_based_on_ipv6_peer(self):
        vpnservice = mock.Mock()
        vpnservice.external_v4_ip = '10.0.0.99'
        vpnservice.external_v6_ip = '2001::1'
        ipsec_sitecon = {'id': FAKE_CONN_ID, 'peer_address': '2001::5'}
        ip_to_use = self.driver.get_external_ip_based_on_peer(vpnservice,
                                                              ipsec_sitecon)
        self.assertEqual('2001::1', ip_to_use)

    def test_get_ipv4_gw_ip(self):
        vpnservice = mock.Mock()
        vpnservice.router.gw_port = {'fixed_ips':
                                     [{'ip_address': '10.0.0.99'}]}
        v4_ip, v6_ip = self.driver._get_gateway_ips(vpnservice.router)
        self.assertEqual('10.0.0.99', v4_ip)
        self.assertIsNone(v6_ip)

    def test_get_ipv6_gw_ip(self):
        vpnservice = mock.Mock()
        vpnservice.router.gw_port = {'fixed_ips': [{'ip_address': '2001::1'}]}
        v4_ip, v6_ip = self.driver._get_gateway_ips(vpnservice.router)
        self.assertIsNone(v4_ip)
        self.assertEqual('2001::1', v6_ip)

    def test_get_both_gw_ips(self):
        vpnservice = mock.Mock()
        vpnservice.router.gw_port = {'fixed_ips': [{'ip_address': '10.0.0.99'},
                                                   {'ip_address': '2001::1'}]}
        v4_ip, v6_ip = self.driver._get_gateway_ips(vpnservice.router)
        self.assertEqual('10.0.0.99', v4_ip)
        self.assertEqual('2001::1', v6_ip)

    def test_use_first_gw_ips_when_multiples(self):
        vpnservice = mock.Mock()
        vpnservice.router.gw_port = {'fixed_ips': [{'ip_address': '10.0.0.99'},
                                                   {'ip_address': '20.0.0.99'},
                                                   {'ip_address': '2001::1'},
                                                   {'ip_address': 'fd00::4'}]}
        v4_ip, v6_ip = self.driver._get_gateway_ips(vpnservice.router)
        self.assertEqual('10.0.0.99', v4_ip)
        self.assertEqual('2001::1', v6_ip)

    def test_store_gw_ips_on_service_create(self):
        vpnservice = mock.Mock()
        self.svc_plugin._get_vpnservice.return_value = vpnservice
        vpnservice.router.gw_port = {'fixed_ips': [{'ip_address': '10.0.0.99'},
                                                   {'ip_address': '2001::1'}]}
        ctxt = n_ctx.Context('', 'somebody')
        vpnservice_dict = {'id': FAKE_SERVICE_ID,
                           'router_id': FAKE_ROUTER_ID}
        self.driver.create_vpnservice(ctxt, vpnservice_dict)
        self.svc_plugin.set_external_tunnel_ips.assert_called_once_with(
            ctxt, FAKE_SERVICE_ID, v4_ip='10.0.0.99', v6_ip='2001::1')

    def test_validate_ipsec_policy(self):
        # Validate IPsec Policy transform_protocol
        ipsec_policy = {'transform_protocol': 'ah-esp'}
        self.assertRaises(ipsec_validator.IpsecValidationFailure,
                          self.validator.validate_ipsec_policy,
                          self.context, ipsec_policy)



# Copyright 2025 SysEleven GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib.api.definitions import vpn_aes_ctr
from neutron_lib.api import extensions
from neutron_lib.plugins import constants as nconstants

from neutron.api.v2 import resource_helper


class Vpn_aes_ctr(extensions.APIExtensionDescriptor):
    api_definition = vpn_aes_ctr

    @classmethod
    def get_resources(cls):
        special_mappings = {'ikepolicies': 'ikepolicy',
                            'ipsecpolicies': 'ipsecpolicy'}
        plural_mappings = resource_helper.build_plural_mappings(
            special_mappings, vpn_aes_ctr.RESOURCE_ATTRIBUTE_MAP)
        return resource_helper.build_resource_info(
            plural_mappings,
            vpn_aes_ctr.RESOURCE_ATTRIBUTE_MAP,
            nconstants.VPN,
            register_quota=True,
            translate_name=True)



# Copyright 2017 Eayun, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add flavor id to vpnservices

Revision ID: 95601446dbcc
Revises: 38893903cbde
Create Date: 2017-04-10 10:14:41.724811

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

# revision identifiers, used by Alembic.
revision = '95601446dbcc'
down_revision = '38893903cbde'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.PIKE, migration.QUEENS,
                     migration.ROCKY, migration.STEIN,
                     migration.TRAIN, migration.USSURI]


def upgrade():
    op.add_column('vpnservices',
                  sa.Column('flavor_id', sa.String(length=36), nullable=True))
    op.create_foreign_key('fk_vpnservices_flavors_id',
                          'vpnservices', 'flavors',
                          ['flavor_id'], ['id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import functools
from unittest import mock

import netaddr
from neutron.agent.common import ovs_lib
from neutron.agent.l3 import l3_agent_extensions_manager as ext_manager
from neutron.agent.l3 import namespaces as n_namespaces
from neutron.agent.l3 import router_info
from neutron.agent import l3_agent as l3_agent_main
from neutron.agent.linux import external_process
from neutron.agent.linux import interface
from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as linux_utils
from neutron.common import config as common_config
from neutron.common import utils as common_utils
from neutron.conf.agent import common as agent_config
from neutron.conf import common as conf_common
from neutron.services.provider_configuration import serviceprovider_opts
from neutron.tests.common import l3_test_common
from neutron.tests.common import net_helpers
from neutron.tests.functional.agent.l3 import framework
from neutron_lib import constants
from neutron_lib.utils import net as n_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import testtools

from neutron_vpnaas.services.vpn import agent as vpn_agent
from neutron_vpnaas.services.vpn.agent import vpn_agent_opts
from neutron_vpnaas.services.vpn.device_drivers import ipsec


_uuid = uuidutils.generate_uuid
FAKE_IKE_POLICY = {
    'auth_algorithm': 'sha1',
    "ike_version": "v1",
    'encryption_algorithm': 'aes-128',
    'pfs': 'group5',
    'phase1_negotiation_mode': 'main',
    'lifetime_units': 'seconds',
    'lifetime_value': 3600
}

FAKE_IPSEC_POLICY = {
    "encapsulation_mode": "tunnel",
    "encryption_algorithm": "aes-128",
    "pfs": "group5",
    "lifetime_units": "seconds",
    "lifetime_value": 3600,
    "transform_protocol": "esp",
    "auth_algorithm": "sha1",
}

FAKE_VPN_SERVICE = {
    "id": _uuid(),
    "router_id": _uuid(),
    "status": constants.PENDING_CREATE,
    "admin_state_up": True,
    'external_ip': "172.24.4.8"
}

FAKE_IPSEC_CONNECTION = {
    "vpnservice_id": _uuid(),
    "status": "PENDING_CREATE",
    "psk": "969022489",
    "initiator": "bi-directional",
    "admin_state_up": True,
    "auth_mode": "psk",
    'external_ip': "172.24.4.8",
    "peer_cidrs": ["10.100.255.224/28"],
    "mtu": 1500,
    "dpd_action": "hold",
    "dpd_interval": 30,
    "dpd_timeout": 120,
    "route_mode": "static",
    "ikepolicy": FAKE_IKE_POLICY,
    "ipsecpolicy": FAKE_IPSEC_POLICY,
    "peer_address": "172.24.4.8",
    "peer_id": "172.24.4.8",
    "id": _uuid()
}


FAKE_IKE_POLICY_SHA256 = {
    'auth_algorithm': 'sha256',
    "ike_version": "v1",
    'encryption_algorithm': 'aes-128',
    'pfs': 'group5',
    'phase1_negotiation_mode': 'main',
    'lifetime_units': 'seconds',
    'lifetime_value': 3600
}

FAKE_IPSEC_POLICY_SHA256 = {
    "encapsulation_mode": "tunnel",
    "encryption_algorithm": "aes-128",
    "pfs": "group5",
    "lifetime_units": "seconds",
    "lifetime_value": 3600,
    "transform_protocol": "esp",
    "auth_algorithm": "sha256",
}

FAKE_IPSEC_CONNECTION_SHA256 = {
    "vpnservice_id": _uuid(),
    "status": "PENDING_CREATE",
    "psk": "969022489",
    "initiator": "bi-directional",
    "admin_state_up": True,
    "auth_mode": "psk",
    'external_ip': "172.24.4.8",
    "peer_cidrs": ["10.100.255.224/28"],
    "mtu": 1500,
    "dpd_action": "hold",
    "dpd_interval": 30,
    "dpd_timeout": 120,
    "route_mode": "static",
    "ikepolicy": FAKE_IKE_POLICY_SHA256,
    "ipsecpolicy": FAKE_IPSEC_POLICY_SHA256,
    "peer_address": "172.24.4.8",
    "peer_id": "172.24.4.8",
    "id": _uuid()
}

PUBLIC_NET = netaddr.IPNetwork('19.4.4.0/24')
PRIVATE_NET = netaddr.IPNetwork('35.4.0.0/16')
FAKE_PUBLIC_SUBNET_ID = _uuid()
FAKE_PRIVATE_SUBNET_ID = _uuid()

FAKE_ROUTER = {
    'enable_snat': True,
    'gw_port': {
        'network_id': _uuid(),
        'subnets': [
            {
                'cidr': str(PUBLIC_NET),
                'gateway_ip': str(PUBLIC_NET[1]),
                'id': FAKE_PUBLIC_SUBNET_ID
            }
        ],
        'fixed_ips': [
            {
                'subnet_id': FAKE_PUBLIC_SUBNET_ID,
                'prefixlen': PUBLIC_NET.prefixlen,
            }
        ],
    },
    'distributed': False,
    '_floatingips': [],
    'routes': []
}

# It's a long name.
NON_ASCII_VPNSERVICE_NAME = '\u9577\u3044\u540d\u524d\u3067\u3059'
# I'm doing very well.
NON_ASCII_PSK = '\u00e7a va tr\u00e9s bien'


def get_ovs_bridge(br_name):
    return ovs_lib.OVSBridge(br_name)


Vm = collections.namedtuple('Vm', ['namespace', 'port_ip'])


class SiteInfo:

    """Holds info on the router, ports, service, and connection."""

    def __init__(self, public_net, private_nets):
        self.public_net = public_net
        self.private_nets = private_nets
        self.generate_router_info()
        self._prepare_vpn_service_info()

    def _get_random_mac(self):
        mac_base = cfg.CONF.base_mac.split(':')
        return n_utils.get_random_mac(mac_base)

    def _generate_private_interface_for_router(self, subnet):
        subnet_id = _uuid()
        return {
            'id': _uuid(),
            'admin_state_up': True,
            'network_id': _uuid(),
            'mtu': 1500,
            'mac_address': self._get_random_mac(),
            'subnets': [
                {
                    'ipv6_ra_mode': None,
                    'cidr': str(subnet),
                    'gateway_ip': str(subnet[1]),
                    'id': subnet_id,
                    'ipv6_address_mode': None
                }
            ],
            'fixed_ips': [
                {
                    'subnet_id': subnet_id,
                    'prefixlen': 24,
                    'ip_address': str(subnet[4])
                }
            ]
        }

    def generate_router_info(self):
        self.info = copy.deepcopy(FAKE_ROUTER)
        self.info['id'] = _uuid()
        self.info['project_id'] = _uuid()
        self.info['_interfaces'] = [
            self._generate_private_interface_for_router(subnet)
            for subnet in self.private_nets]
        self.info['gw_port']['id'] = _uuid()
        self.info['gw_port']['fixed_ips'][0]['ip_address'] = str(
            self.public_net)
        self.info['gw_port']['mac_address'] = self._get_random_mac()
        self.info['ha'] = False

    def _prepare_vpn_service_info(self):
        self.vpn_service = copy.deepcopy(FAKE_VPN_SERVICE)
        self.vpn_service.update({'id': _uuid(),
                                 'router_id': self.info['id'],
                                 'external_ip': str(self.public_net)})

    def prepare_ipsec_conn_info(self, peer, connection=FAKE_IPSEC_CONNECTION,
                                local_id=None, peer_id=None):
        ipsec_connection = copy.deepcopy(connection)
        local_cidrs = [str(s) for s in self.private_nets]
        peer_cidrs = [str(s) for s in peer.private_nets]
        ipsec_connection.update({
            'id': _uuid(),
            'vpnservice_id': self.vpn_service['id'],
            'external_ip': self.vpn_service['external_ip'],
            'peer_cidrs': peer_cidrs,
            'peer_address': peer.vpn_service['external_ip'],
            'peer_id': peer.vpn_service['external_ip'],
            'local_cidrs': local_cidrs,
            'local_ip_vers': 4
        })
        if local_id:
            ipsec_connection['local_id'] = local_id
        if peer_id:
            ipsec_connection['peer_id'] = peer_id
        self.vpn_service['ipsec_site_connections'] = [ipsec_connection]


class SiteInfoWithHaRouter(SiteInfo):

    def __init__(self, public_net, private_nets, host, failover_host):
        self.host = host
        self.failover_host = failover_host
        self.get_ns_name = mock.patch.object(n_namespaces.RouterNamespace,
                                             '_get_ns_name').start()
        super().__init__(public_net, private_nets)

    def generate_router_info(self):
        super().generate_router_info()
        self.info['ha'] = True
        self.info['ha_vr_id'] = 1
        self.info[constants.HA_INTERFACE_KEY] = (
            l3_test_common.get_ha_interface())
        # Mock router namespace name, for when router is created
        self.get_ns_name.return_value = "qrouter-{}-{}".format(
            self.info['id'], self.host)

    def generate_backup_router_info(self):
        # Clone router info, using different HA interface (using same ID)
        info = copy.deepcopy(self.info)
        info[constants.HA_INTERFACE_KEY] = (
            l3_test_common.get_ha_interface(ip='169.254.192.2',
                                            mac='22:22:22:22:22:22'))
        # Mock router namespace name, for when router is created
        self.get_ns_name.return_value = "qrouter-{}-{}".format(
            info['id'], self.failover_host)
        return info


class TestIPSecBase(framework.L3AgentTestFramework):
    NESTED_NAMESPACE_SEPARATOR = '@'

    def setUp(self):
        super().setUp()
        common_config.register_common_config_options()
        mock.patch('neutron_vpnaas.services.vpn.device_drivers.ipsec.'
                   'IPsecVpnDriverApi').start()
        # avoid report_status running periodically
        mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall').start()
        # Both the vpn agents try to use execute_rootwrap_daemon's socket
        # simultaneously during test cleanup, but execute_rootwrap_daemon has
        # limitations with simultaneous reads. So avoid using
        # root_helper_daemon and instead use root_helper
        # https://bugs.launchpad.net/neutron/+bug/1482622
        cfg.CONF.set_override('root_helper_daemon', None, group='AGENT')

        # Mock the method below because it causes Exception:
        #   RuntimeError: Second simultaneous read on fileno 5 detected.
        #   Unless you really know what you're doing, make sure that only
        #   one greenthread can read any particular socket.  Consider using
        #   a pools.Pool. If you do know what you're doing and want to disable
        #   this error, call eventlet.debug.hub_prevent_multiple_readers(False)
        # Can reproduce the exception in the test only
        ip_lib.send_ip_addr_adv_notif = mock.Mock()

        self.vpn_agent = vpn_agent.L3WithVPNaaS(self.conf)
        self.driver = self.vpn_agent.device_drivers[0]
        self.driver.agent_rpc.get_vpn_services_on_host = mock.Mock(
            return_value=[])
        self.driver.report_status = mock.Mock()

        self.private_nets = list(PRIVATE_NET.subnet(24))

    def _connect_agents(self, agent1, agent2):
        """Simulate both agents in the same host.

         For packet flow between resources connected to these two agents,
         agent's ovs bridges are connected through patch ports.
        """
        br_int_1 = get_ovs_bridge(agent1.conf.OVS.integration_bridge)
        br_int_2 = get_ovs_bridge(agent2.conf.OVS.integration_bridge)
        net_helpers.create_patch_ports(br_int_1, br_int_2)

    def _get_config_opts(self):
        """Register default config options"""
        config = cfg.ConfigOpts()
        config.register_opts(conf_common.core_opts)
        config.register_opts(conf_common.core_cli_opts)
        config.register_opts(serviceprovider_opts, 'service_providers')
        config.register_opts(vpn_agent_opts, 'vpnagent')
        config.register_opts(ipsec.ipsec_opts, 'ipsec')
        config.register_opts(ipsec.openswan_opts, 'openswan')

        logging.register_options(config)
        agent_config.register_process_monitor_opts(config)
        ext_manager.register_opts(config)
        return config

    def _configure_agent(self, host):
        """Override specific config options"""
        config = self._get_config_opts()
        l3_agent_main.register_opts(config)
        cfg.CONF.set_override('debug', True)
        agent_config.setup_logging()
        config.set_override('extensions', ['vpnaas'], 'agent')
        config.set_override(
            'interface_driver',
            'neutron.agent.linux.interface.OVSInterfaceDriver')

        br_int = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        config.set_override('integration_bridge', br_int.br_name, 'OVS')

        temp_dir = self.get_new_temp_dir()
        get_temp_file_path = functools.partial(self.get_temp_file_path,
                                               root=temp_dir)
        config.set_override('state_path', temp_dir.path)
        config.set_override('metadata_proxy_socket',
                            get_temp_file_path('metadata_proxy'))
        config.set_override('ha_confs_path',
                            get_temp_file_path('ha_confs'))
        config.set_override('external_pids',
                            get_temp_file_path('external/pids'))
        config.set_override('host', host)
        ipsec_config_base_dir = '{}/{}'.format(temp_dir.path, 'ipsec')
        config.set_override('config_base_dir',
                            ipsec_config_base_dir, group='ipsec')

        # Assign ip address to br-int port because it is a gateway
        ex_port = ip_lib.IPDevice(br_int.br_name)
        ex_port.addr.add(str(PUBLIC_NET[1]))

        return config

    def _setup_failover_agent(self):
        self.failover_agent = self._configure_agent('agent2')
        self._connect_agents(self.vpn_agent, self.failover_agent)
        self.failover_driver = self.failover_agent.device_drivers[0]
        self.failover_driver.agent_rpc.get_vpn_services_on_host = (
            mock.Mock(return_value=[]))
        self.failover_driver.report_status = mock.Mock()

    def create_router(self, agent, info):
        """Create router for agent from router info."""
        self.addCleanup(agent._safe_router_removed, info['id'])

        # Generate unique internal and external router device names using the
        # agent's hostname. This is to allow multiple HA router replicas to
        # co-exist on the same machine, otherwise they'd all use the same
        # device names and OVS would freak out(OVS won't allow a port with
        # same name connected to two bridges).
        def _append_suffix(dev_name):
            # If dev_name = 'xyz123' and the suffix is 'agent2' then the result
            # will be 'xy-nt2'
            return f"{dev_name[:-4]}-{agent.host[-3:]}"

        def get_internal_device_name(port_id):
            return _append_suffix(
                (n_namespaces.INTERNAL_DEV_PREFIX + port_id)
                [:interface.LinuxInterfaceDriver.DEV_NAME_LEN])

        def get_external_device_name(port_id):
            return _append_suffix(
                (n_namespaces.EXTERNAL_DEV_PREFIX + port_id)
                [:interface.LinuxInterfaceDriver.DEV_NAME_LEN])

        mock_get_internal_device_name = mock.patch.object(
            router_info.RouterInfo, 'get_internal_device_name').start()
        mock_get_internal_device_name.side_effect = get_internal_device_name
        mock_get_external_device_name = mock.patch.object(
            router_info.RouterInfo, 'get_external_device_name').start()
        mock_get_external_device_name.side_effect = get_external_device_name

        # NOTE(huntxu): with commit 88f5e11d8bf, neutron plugs new ports as
        # dead vlan(4095). During functional tests, all the ports are untagged.
        # So need to remove such tag during functional testing.
        original_plug_new = interface.OVSInterfaceDriver.plug_new

        def plug_new(self, *args, **kwargs):
            original_plug_new(self, *args, **kwargs)
            bridge = (kwargs.get('bridge') or args[4] or
                      self.conf.OVS.integration_bridge)
            device_name = kwargs.get('device_name') or args[2]
            ovsbr = ovs_lib.OVSBridge(bridge)
            ovsbr.clear_db_attribute('Port', device_name, 'tag')

        with mock.patch(
            'neutron.agent.linux.interface.OVSInterfaceDriver.plug_new',
            autospec=True
        ) as ovs_plug_new:
            ovs_plug_new.side_effect = plug_new
            agent._process_added_router(info)

        return agent.router_info[info['id']]

    def _port_first_ip_cidr(self, port):
        fixed_ip = port['fixed_ips'][0]
        return common_utils.ip_to_cidr(fixed_ip['ip_address'],
                                       fixed_ip['prefixlen'])

    def create_ports_for(self, site):
        """Creates namespaces and ports for simulated VM.

        There will be a unique namespace for each port, which is representing
        a VM for the test.
        """
        bridge = get_ovs_bridge(self.vpn_agent.conf.OVS.integration_bridge)
        site.vm = []
        for internal_port in site.router.internal_ports:
            router_ip_cidr = self._port_first_ip_cidr(internal_port)
            port_ip_cidr = net_helpers.increment_ip_cidr(router_ip_cidr, 1)
            client_ns = self.useFixture(
                net_helpers.NamespaceFixture()).ip_wrapper
            namespace = client_ns.namespace
            port = self.useFixture(
                net_helpers.OVSPortFixture(bridge, namespace)).port
            port.addr.add(port_ip_cidr)
            port.route.add_gateway(router_ip_cidr.partition('/')[0])
            site.vm.append(Vm(namespace, port_ip_cidr.partition('/')[0]))

    def create_site(self, public_net, private_nets, l3ha=False):
        """Build router(s), namespaces, and ports for a site.

        For HA, we'll create a backup router and wait for both routers
        to be ready, so that we can test pings after failover.
        """
        if l3ha:
            site = SiteInfoWithHaRouter(public_net, private_nets,
                                        self.agent.host,
                                        self.failover_agent.host)
        else:
            site = SiteInfo(public_net, private_nets)

        site.router = self.create_router(self.agent, site.info)
        if l3ha:
            backup_info = site.generate_backup_router_info()
            site.backup_router = self.create_router(self.failover_agent,
                                                    backup_info)
            linux_utils.wait_until_true(
                lambda: site.router.ha_state in ('master', 'primary'))
            linux_utils.wait_until_true(
                lambda: site.backup_router.ha_state == 'backup')

        self.create_ports_for(site)
        return site

    def prepare_ipsec_site_connections(self, site1, site2):
        """Builds info for connections in both directions in prep for sync."""
        site1.prepare_ipsec_conn_info(site2)
        site2.prepare_ipsec_conn_info(site1)

    def prepare_ipsec_site_connections_sha256(self, site1, site2):
        """Builds info for connections in both directions in prep for sync."""
        site1.prepare_ipsec_conn_info(site2,
                                      FAKE_IPSEC_CONNECTION_SHA256)
        site2.prepare_ipsec_conn_info(site1,
                                      FAKE_IPSEC_CONNECTION_SHA256)

    def prepare_ipsec_site_connections_local_id(self, site1, site2):
        """Builds info for connections in both directions in prep for sync."""
        site1.prepare_ipsec_conn_info(site2, local_id='@site1.com',
                                      peer_id='@site2.com')
        site2.prepare_ipsec_conn_info(site1, local_id='@site2.com',
                                      peer_id='@site1.com')

    def sync_to_create_ipsec_connections(self, site1, site2):
        """Perform a sync, so that connections are created."""
        # Provide service info to sync
        self.driver.agent_rpc.get_vpn_services_on_host = mock.Mock(
            return_value=[site1.vpn_service, site2.vpn_service])

        local_router_id = site1.router.router_id
        peer_router_id = site2.router.router_id
        self.driver.sync(mock.Mock(), [site1.router,
                                       site2.router])
        self.agent._process_updated_router(site1.router.router)
        self.agent._process_updated_router(site2.router.router)
        self.addCleanup(self.driver._delete_vpn_processes,
                        [local_router_id, peer_router_id], [])

    def sync_failover_agent(self, site):
        """Perform a sync on failover agent associated w/backup router."""
        self.failover_driver.agent_rpc.get_vpn_services_on_host = mock.Mock(
            return_value=[site.vpn_service])
        self.failover_driver.sync(mock.Mock(), [site.router])

    def check_ping(self, from_site, to_site, instance=0, success=True):
        if success:
            net_helpers.assert_ping(from_site.vm[instance].namespace,
                                    to_site.vm[instance].port_ip,
                                    timeout=8, count=4)
        else:
            net_helpers.assert_no_ping(from_site.vm[instance].namespace,
                                       to_site.vm[instance].port_ip,
                                       timeout=8, count=4)

    def _failover_ha_router(self, router1, router2):
        """Cause a failover of HA router.

        Fail the agent1's HA router. Agent1's HA router will transition
        to backup and agent2's HA router will become master. Wait for
        the failover to complete.
        """
        device_name = router1.get_ha_device_name()
        ha_device = ip_lib.IPDevice(device_name, router1.ns_name)
        ha_device.link.set_down()
        linux_utils.wait_until_true(
            lambda: router2.ha_state in ('master', 'primary'))
        linux_utils.wait_until_true(lambda: router1.ha_state == 'backup')

    def _ipsec_process_exists(self, conf, router, pid_files):
        """Check if *Swan process has started up."""
        for pid_file in pid_files:
            pm = external_process.ProcessManager(
                conf,
                "ipsec",
                router.ns_name, pid_file=pid_file)
            if pm.active:
                break
        return pm.active

    def _wait_for_ipsec_startup(self, router, driver, conf, should_run=True):
        """Wait for new IPSec process on failover agent to start up."""
        # check for both strongswan and openswan processes
        path = driver.processes[router.router_id].config_dir
        pid_files = ['%s/var/run/charon.pid' % path,
                     '%s/var/run/pluto.pid' % path]
        linux_utils.wait_until_true(
            lambda: should_run == self._ipsec_process_exists(
                conf, router, pid_files))

    @staticmethod
    def _update_vpnservice(site, **kwargs):
        site.vpn_service.update(kwargs)

    @staticmethod
    def _update_ipsec_connection(site, **kwargs):
        ipsec_connection = site.vpn_service['ipsec_site_connections'][0]
        ipsec_connection.update(kwargs)


class TestIPSecScenario(TestIPSecBase):

    @testtools.skip('bug/1598466')
    def test_single_ipsec_connection(self):
        site1 = self.create_site(PUBLIC_NET[4], [self.private_nets[1]])
        site2 = self.create_site(PUBLIC_NET[5], [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    @testtools.skip('bug/1598466')
    def test_single_ipsec_connection_sha256(self):
        site1 = self.create_site(PUBLIC_NET[4], [self.private_nets[1]])
        site2 = self.create_site(PUBLIC_NET[5], [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections_sha256(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    @testtools.skip('bug/1598466')
    def test_single_ipsec_connection_local_id(self):
        site1 = self.create_site(PUBLIC_NET[4], [self.private_nets[1]])
        site2 = self.create_site(PUBLIC_NET[5], [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections_local_id(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    @testtools.skip('bug/1598466')
    def test_ipsec_site_connections_with_mulitple_subnets(self):
        """Check with a pair of subnets on each end of connection."""
        site1 = self.create_site(PUBLIC_NET[4], self.private_nets[1:3])
        site2 = self.create_site(PUBLIC_NET[5], self.private_nets[3:5])

        # Just check from each VM, not every combination
        for i in [0, 1]:
            self.check_ping(site1, site2, instance=i, success=False)
            self.check_ping(site2, site1, instance=i, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        for i in [0, 1]:
            self.check_ping(site1, site2, instance=i)
            self.check_ping(site2, site1, instance=i)

    @testtools.skip('bug/1598466')
    def test_ipsec_site_connections_with_l3ha_routers(self):
        """Test ipsec site connection with HA routers.
        This test creates two agents. First agent will have Legacy and HA
        routers. Second agent will host only HA router. We setup ipsec
        connection between legacy and HA router.

        When HA router is created, agent1 will have master router and
        agent2 will have backup router. Ipsec connection will be established
        between legacy router and agent1's master HA router.

        Then we fail the agent1's master HA router. Agent1's HA router will
        transition to backup and agent2's HA router will become master.
        Now ipsec connection will be established between legacy router and
        agent2's master HA router
        """

        self._setup_failover_agent()

        site1 = self.create_site(PUBLIC_NET[4], [self.private_nets[1]])
        site2 = self.create_site(PUBLIC_NET[5], [self.private_nets[2]],
                                 l3ha=True)

        # No ipsec connection between legacy router and HA routers
        self.check_ping(site1, site2, 0, success=False)
        self.check_ping(site2, site1, 0, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)
        self.sync_failover_agent(site2)

        # Test ipsec connection between legacy router and agent2's HA router
        self.check_ping(site1, site2, 0)
        self.check_ping(site2, site1, 0)

        self._failover_ha_router(site2.router, site2.backup_router)
        self._wait_for_ipsec_startup(site2.backup_router,
                                     self.failover_driver,
                                     self.failover_agent.conf)

        # Test ipsec connection between legacy router and agent2's HA router
        self.check_ping(site1, site2, 0)
        self.check_ping(site2, site1, 0)

    @testtools.skip('bug/1598466')
    def _test_admin_state_up(self, update_method):
        # Create ipsec connection between two sites
        site1 = self.create_site(PUBLIC_NET[4], [self.private_nets[1]])
        site2 = self.create_site(PUBLIC_NET[5], [self.private_nets[2]])

        self.prepare_ipsec_site_connections(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

        # Disable resource on one of the sites and check that
        # ping no longer passes.
        update_method(site1, admin_state_up=False)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2, 0, success=False)
        self.check_ping(site2, site1, 0, success=False)

        # Validate that ipsec process for the disabled site was terminated.
        self._wait_for_ipsec_startup(site1.router, self.driver,
                                     self.vpn_agent.conf,
                                     should_run=False)

        # Change admin_state_up of the disabled resource back to True and
        # check that everything works again.
        update_method(site1, admin_state_up=True)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    @testtools.skip('bug/1598466')
    def test_ipsec_site_connections_update_admin_state_up(self):
        """Test updating admin_state_up of ipsec site connections."""

        self._test_admin_state_up(self._update_ipsec_connection)

    @testtools.skip('bug/1598466')
    def test_vpnservice_update_admin_state_up(self):
        """Test updating admin_state_up of a vpn service."""

        self._test_admin_state_up(self._update_vpnservice)



# Copyright 2025 SysEleven GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""add more ciphers

Revision ID: b18aab30fddc
Revises: 22e0145ac80b
Create Date: 2023-10-11 15:40:27.845720

"""

# revision identifiers, used by Alembic.
revision = 'b18aab30fddc'
down_revision = '22e0145ac80b'


new_auth = sa.Enum(
    'sha1', 'sha256', 'sha384', 'sha512', 'aes-xcbc', 'aes-cmac',
    name="vpn_pfs",
)

new_enc = sa.Enum(
    '3des',
    'aes-128', 'aes-192', 'aes-256',
    'aes-128-ctr', 'aes-192-ctr', 'aes-256-ctr',
    'aes-128-ccm-8', 'aes-192-ccm-8', 'aes-256-ccm-8',
    'aes-128-ccm-12', 'aes-192-ccm-12', 'aes-256-ccm-12',
    'aes-128-ccm-16', 'aes-192-ccm-16', 'aes-256-ccm-16',
    'aes-128-gcm-8', 'aes-192-gcm-8', 'aes-256-gcm-8',
    'aes-128-gcm-12', 'aes-192-gcm-12', 'aes-256-gcm-12',
    'aes-128-gcm-16', 'aes-192-gcm-16', 'aes-256-gcm-16',
    name="vpn_encrypt_algorithms",
)

new_pfs = sa.Enum(
    'group2', 'group5', 'group14', 'group15',
    'group16', 'group17', 'group18', 'group19', 'group20', 'group21',
    'group22', 'group23', 'group24', 'group25', 'group26', 'group27',
    'group28', 'group29', 'group30', 'group31',
    name="vpn_pfs",
)


def upgrade():
    op.alter_column('ikepolicies', 'pfs', type_=new_pfs,
                    existing_nullable=False)
    op.alter_column('ikepolicies', 'auth_algorithm', type_=new_auth,
                    existing_nullable=False)
    op.alter_column('ikepolicies', 'encryption_algorithm', type_=new_enc,
                    existing_nullable=False)

    op.alter_column('ipsecpolicies', 'pfs', type_=new_pfs,
                    existing_nullable=False)
    op.alter_column('ipsecpolicies', 'auth_algorithm', type_=new_auth,
                    existing_nullable=False)
    op.alter_column('ipsecpolicies', 'encryption_algorithm', type_=new_enc,
                    existing_nullable=False)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from rally.common import logging
from rally.task import scenario
from rally.task import types

import vpn_base
LOG = logging.getLogger(__name__)


class TestVpnBasicScenario(vpn_base.VpnBase):
    """Rally scenarios for VPNaaS"""

    @types.convert(image={"type": "glance_image"},
                   flavor={"type": "nova_flavor"})
    @scenario.configure()
    def create_and_delete_vpn_connection(self, **kwargs):
        """Basic VPN connectivity scenario.

        1. Create 2 private networks, subnets and routers
        2. Create public network, subnets and GW IPs on routers, if not present
        3. Execute ip netns command and get the snat and qrouter namespaces
           (assuming we use DVR)
        4. Verify that there is a route between the router gateways by pinging
           each other from their snat namespaces
        5. Add security group rules for SSH and ICMP
        6. Start a nova instance in each of the private networks
        7. Create IKE and IPSEC policies
        8. Create VPN service at each of the routers
        9. Create IPSEC site connections at both endpoints
        10. Verify that the ipsec-site-connection is ACTIVE (takes upto 30secs)
        11. To verify the vpn connectivity, get into the peer router's snat
            namespace and start a tcpdump at the qg-xxxx interface
        12. SSH into the nova instance from the local qrouter namespace
            and try to ping the nova instance on the peer network.
        14. Verify that the captured packets are encapsulated and encrypted.
        15. Verify the connectivity in the reverse direction following the
            steps 11 through 13
        16. Submit a request to delete all the resources
        """

        try:
            self.setup(**kwargs)
            self.create_networks(**kwargs)
            self.create_servers(**kwargs)
            self.check_route()
            self.ike_policy = self._create_ike_policy(**kwargs)
            self.ipsec_policy = self._create_ipsec_policy(**kwargs)
            self.create_vpn_services()
            self.create_ipsec_site_connections(**kwargs)
            self.assert_statuses(final_status='ACTIVE', **kwargs)
            self.verify_vpn_connectivity(**kwargs)
            LOG.info("VPN CONNECTIVITY TEST PASSED!")

        finally:
            self.cleanup()



# Copyright (c) 2015 Canonical, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from neutron.agent.linux import utils
from neutron.conf.agent import common as config
from neutron.tests.common import net_helpers
from neutron.tests.functional import base

WRAPPER_SCRIPT = 'neutron-vpn-netns-wrapper'
STATUS_PATTERN = re.compile('Command:.*ip.*addr.*show.*Exit code: 0')


class TestNetnsWrapper(base.BaseSudoTestCase):

    def setUp(self):
        super().setUp()
        config.setup_logging()
        self.fake_ns = 'func-8f1b728c-6eca-4042-9b6b-6ef66ab9352a'
        self.mount_paths = ('--mount_paths=/etc:/var/lib/neutron'
                            '/vpnaas/%(ns)s/etc,/var/run:/var/lib'
                            '/neutron/vpnaas/%(ns)s/var/run')
        self.fake_pth = self.mount_paths % {'ns': self.fake_ns}

    def test_netns_wrap_success(self):
        client_ns = self.useFixture(net_helpers.NamespaceFixture()).ip_wrapper
        ns = client_ns.namespace
        pth = self.mount_paths % {'ns': ns}
        cmd = WRAPPER_SCRIPT, pth, '--cmd=ip,addr,show'
        output = client_ns.netns.execute(cmd)
        self.assertTrue(STATUS_PATTERN.search(output))

    def test_netns_wrap_fail_without_netns(self):
        cmd = [WRAPPER_SCRIPT, self.fake_pth,
               '--cmd=ip,addr,show']
        self.assertRaises(RuntimeError, utils.execute, cmd=cmd,
                          run_as_root=True)

    def test_netns_wrap_unauthorized_command(self):
        cmd = [WRAPPER_SCRIPT, self.fake_pth,
               '--cmd=nofiltercommand']
        self.assertRaises(RuntimeError, utils.execute, cmd=cmd,
                          run_as_root=True)



# Copyright 2020, SysEleven GbmH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from neutron_lib import context as n_ctx
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils

from neutron_vpnaas.services.vpn.service_drivers import ipsec_validator
from neutron_vpnaas.services.vpn.service_drivers \
    import ovn_ipsec as ipsec_driver
from neutron_vpnaas.tests import base


_uuid = uuidutils.generate_uuid

FAKE_HOST = 'fake_host'
FAKE_TENANT_ID = 'tenant1'
FAKE_ROUTER_ID = _uuid()
FAKE_TRANSIT_IP_ADDRESS = '169.254.0.2'

FAKE_VPNSERVICE_1 = {
    'id': _uuid(),
    'router_id': FAKE_ROUTER_ID,
    'tenant_id': FAKE_TENANT_ID
}

FAKE_VPNSERVICE_2 = {
    'id': _uuid(),
    'router_id': FAKE_ROUTER_ID,
    'tenant_id': FAKE_TENANT_ID
}

FAKE_VPN_CONNECTION_1 = {
    'vpnservice_id': FAKE_VPNSERVICE_1['id']
}


class FakeSqlQueryObject(dict):
    """To fake SqlAlchemy query object and access keys as attributes."""

    def __init__(self, **entries):
        self.__dict__.update(entries)
        super().__init__(**entries)


class FakeGatewayDB:
    def __init__(self):
        self.gateways_by_router = {}
        self.gateways_by_id = {}

    def create_gateway(self, context, gateway):
        info = gateway['gateway']
        fake_gw = {
            'id': _uuid(),
            'status': 'PENDING_CREATE',
            'external_fixed_ips': [{'subnet_id': '1',
                                    'ip_address': '10.2.3.4'}],
            **info
        }
        self.gateways_by_router[info['router_id']] = fake_gw
        self.gateways_by_id[fake_gw['id']] = fake_gw
        return fake_gw

    def update_gateway(self, context, gateway_id, gateway):
        self.gateways_by_id[gateway_id].update(**gateway['gateway'])

    def delete_gateway(self, context, gateway_id):
        fake_gw = self.gateways_by_id.pop(gateway_id, None)
        if fake_gw:
            self.gateways_by_router.pop(fake_gw['router_id'])
        return 1 if fake_gw else 0

    def get_vpn_gw_dict_by_router_id(self, context, router_id, refresh=False):
        return self.gateways_by_router.get(router_id)


class TestOvnIPsecDriver(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        mock.patch('neutron_lib.rpc.Connection').start()
        self.create_port = \
            mock.patch('neutron_lib.plugins.utils.create_port').start()
        self.create_network = \
            mock.patch('neutron_lib.plugins.utils.create_network').start()
        self.create_subnet = \
            mock.patch('neutron_lib.plugins.utils.create_subnet').start()

        self.create_port.side_effect = lambda pl, c, p: {
            'id': _uuid(),
            'fixed_ips': [{'subnet_id': '1', 'ip_address': '10.1.1.2'}]}
        self.create_network.side_effect = lambda pl, c, n: {'id': _uuid()}
        self.create_subnet.side_effect = lambda pl, c, s: {'id': _uuid()}

        vpn_agent = {'host': FAKE_HOST}

        self.core_plugin = mock.Mock()
        self.core_plugin.get_vpn_agents_hosting_routers.return_value = \
            [vpn_agent]

        directory.add_plugin(constants.CORE, self.core_plugin)

        self._fake_router = FakeSqlQueryObject(
            id=FAKE_ROUTER_ID,
            gw_port=FakeSqlQueryObject(network_id=_uuid())
        )

        self.l3_plugin = mock.Mock()
        self.l3_plugin.get_router.return_value = self._fake_router
        directory.add_plugin(constants.L3, self.l3_plugin)

        self.svc_plugin = mock.Mock()
        self.svc_plugin.get_vpn_agents_hosting_routers.return_value = \
            [vpn_agent]
        self.svc_plugin.schedule_router.return_value = vpn_agent
        self.svc_plugin._get_vpnservice.return_value = FakeSqlQueryObject(
            router_id=FAKE_ROUTER_ID,
            router=self._fake_router
        )
        self.svc_plugin.get_vpn_agent_on_host.return_value = {'id': FAKE_HOST}
        self.svc_plugin.get_vpnservice.return_value = FAKE_VPNSERVICE_1
        self.svc_plugin.get_vpnservice_router_id.return_value = FAKE_ROUTER_ID
        self.driver = ipsec_driver.IPsecOvnVPNDriver(self.svc_plugin)
        self.driver.make_vpnservice_dict = mock.Mock(
            return_value={'id': FAKE_HOST})
        self.validator = ipsec_validator.IpsecVpnValidator(self.driver)
        self.context = n_ctx.get_admin_context()

    def test_create_vpnservice(self):
        mock.patch.object(self.driver.agent_rpc.client, 'cast')
        mock.patch.object(self.driver.agent_rpc.client, 'prepare')
        fake_gw_db = FakeGatewayDB()

        self.svc_plugin.get_vpn_gw_dict_by_router_id.side_effect = \
            fake_gw_db.get_vpn_gw_dict_by_router_id
        self.svc_plugin.create_gateway.side_effect = fake_gw_db.create_gateway
        self.svc_plugin.update_gateway.side_effect = fake_gw_db.update_gateway

        self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE_1)
        self.svc_plugin.create_gateway.assert_called_once()

        # check that the plugin utils create functions were called
        self.create_port.assert_called()
        self.create_network.assert_called_once()
        self.create_subnet.assert_called_once()

        # check that the core plugin create functions were not called directly
        self.core_plugin.create_port.assert_not_called()
        self.core_plugin.create_network.assert_not_called()
        self.core_plugin.create_subnet.assert_not_called()

        self.svc_plugin.reset_mock()

        self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE_2)
        self.svc_plugin.create_gateway.assert_not_called()

    def test_delete_vpnservice(self):
        mock.patch.object(self.driver.agent_rpc.client, 'cast')
        mock.patch.object(self.driver.agent_rpc.client, 'prepare')
        fake_gw_db = FakeGatewayDB()
        self.svc_plugin.get_vpn_gw_dict_by_router_id.side_effect = \
            fake_gw_db.get_vpn_gw_dict_by_router_id
        self.svc_plugin.create_gateway.side_effect = fake_gw_db.create_gateway
        self.svc_plugin.update_gateway.side_effect = fake_gw_db.update_gateway
        self.svc_plugin.delete_gateway.side_effect = fake_gw_db.delete_gateway

        # create 2 VPN services on same router
        self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE_1)
        self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE_2)
        self.svc_plugin.reset_mock()

        # deleting one VPN service must not delete the VPN gateway
        self.svc_plugin.get_vpnservices.return_value = [FAKE_VPNSERVICE_2]
        self.driver.delete_vpnservice(self.context, FAKE_VPNSERVICE_1)
        self.core_plugin.delete_port.assert_not_called()
        self.core_plugin.delete_network.assert_not_called()
        self.core_plugin.delete_subnet.assert_not_called()
        self.svc_plugin.create_gateway.assert_not_called()
        self.svc_plugin.delete_gateway.assert_not_called()

        # deleting last VPN service shall delete the VPN gateway
        self.svc_plugin.get_vpnservices.return_value = []
        self.driver.delete_vpnservice(self.context, FAKE_VPNSERVICE_1)
        self.core_plugin.delete_port.assert_called()
        self.core_plugin.delete_network.assert_called_once()
        self.core_plugin.delete_subnet.assert_called_once()
        self.svc_plugin.create_gateway.assert_not_called()
        self.svc_plugin.delete_gateway.assert_called_once()

    def _test_ipsec_site_connection(self, old_peers, new_peers,
                                    func, args,
                                    expected_add, expected_remove):
        self._fake_router['routes'] = [
            {'destination': peer, 'nexthop': FAKE_TRANSIT_IP_ADDRESS}
            for peer in old_peers
        ]
        transit_port = FakeSqlQueryObject(
            id=_uuid(),
            fixed_ips=[
                {'subnet_id': _uuid(), 'ip_address': FAKE_TRANSIT_IP_ADDRESS}
            ]
        )
        self.svc_plugin.get_vpn_gw_by_router_id.return_value = \
            FakeSqlQueryObject(id=_uuid(),
                               router_id=FAKE_ROUTER_ID,
                               transit_port_id=transit_port.id,
                               transit_port=transit_port)

        self.svc_plugin.get_peer_cidrs_for_router.return_value = new_peers

        # create/update/delete_ipsec_site_connection
        with mock.patch.object(self.driver.agent_rpc.client, 'cast'
                               ) as rpc_mock, \
                mock.patch.object(self.driver.agent_rpc.client, 'prepare'
                                  ) as prepare_mock:
            prepare_mock.return_value = self.driver.agent_rpc.client
            func(self.context, *args)

        prepare_args = {'server': 'fake_host', 'version': '1.0'}
        prepare_mock.assert_called_once_with(**prepare_args)

        # check that agent RPC vpnservice_updated is called
        rpc_mock.assert_called_once_with(self.context, 'vpnservice_updated',
                                         router={'id': FAKE_ROUTER_ID})

        # check that routes were updated
        if expected_add:
            expected_router = {'router': {'routes': [
                {'destination': peer,
                 'nexthop': FAKE_TRANSIT_IP_ADDRESS}
                for peer in expected_add
            ]}}
            self.l3_plugin.add_extraroutes.assert_called_once_with(
                self.context, FAKE_ROUTER_ID, expected_router)
        else:
            self.l3_plugin.add_extraroutes.assert_not_called()

        if expected_remove:
            expected_router = {'router': {'routes': [
                {'destination': peer,
                 'nexthop': FAKE_TRANSIT_IP_ADDRESS}
                for peer in expected_remove
            ]}}
            self.l3_plugin.remove_extraroutes.assert_called_once_with(
                self.context, FAKE_ROUTER_ID, expected_router)
        else:
            self.l3_plugin.remove_extraroutes.assert_not_called()

    def test_create_ipsec_site_connection_1(self):
        old_peers = []
        new_peers = ['192.168.1.0/24']
        expected_add = new_peers
        expected_remove = []
        self._test_ipsec_site_connection(
            old_peers, new_peers,
            self.driver.create_ipsec_site_connection,
            [FAKE_VPN_CONNECTION_1],
            expected_add, expected_remove
        )

    def test_create_ipsec_site_connection_2(self):
        """Test creating a 2nd site connection."""
        old_peers = ['192.168.1.0/24']
        new_peers = ['192.168.1.0/24', '192.168.2.0/24']
        expected_add = ['192.168.2.0/24']
        expected_remove = []
        self._test_ipsec_site_connection(
            old_peers, new_peers,
            self.driver.create_ipsec_site_connection,
            [FAKE_VPN_CONNECTION_1],
            expected_add, expected_remove
        )

    def test_update_ipsec_site_connection(self):
        old_peers = ['192.168.1.0/24']
        new_peers = ['192.168.2.0/24']
        expected_add = new_peers
        expected_remove = old_peers
        self._test_ipsec_site_connection(
            old_peers, new_peers,
            self.driver.update_ipsec_site_connection,
            [FAKE_VPN_CONNECTION_1, FAKE_VPN_CONNECTION_1],
            expected_add, expected_remove
        )

    def test_delete_ipsec_site_connection(self):
        old_peers = ['192.168.1.0/24', '192.168.2.0/24']
        new_peers = ['192.168.2.0/24']
        expected_add = []
        expected_remove = ['192.168.1.0/24']
        self._test_ipsec_site_connection(
            old_peers, new_peers,
            self.driver.delete_ipsec_site_connection,
            [FAKE_VPN_CONNECTION_1],
            expected_add, expected_remove
        )


class TestOvnIPsecCallBackDriver(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        vpn_agent = {'host': FAKE_HOST}
        self.core_plugin = mock.Mock()
        self.core_plugin.get_vpn_agents_hosting_routers.return_value = \
            [vpn_agent]

        directory.add_plugin(constants.CORE, self.core_plugin)
        self._fake_router = FakeSqlQueryObject(
            id=FAKE_ROUTER_ID,
            gw_port=FakeSqlQueryObject(network_id=_uuid())
        )
        self.svc_plugin = mock.Mock()
        self.svc_plugin.get_vpn_agents_hosting_routers.return_value = \
            [vpn_agent]
        self.svc_plugin.schedule_router.return_value = vpn_agent
        self.svc_plugin._get_vpnservice.return_value = FakeSqlQueryObject(
            router_id=FAKE_ROUTER_ID,
            router=self._fake_router
        )
        self.svc_plugin.get_vpn_agent_on_host.return_value = {'id': FAKE_HOST}
        self.svc_plugin.get_vpnservice.return_value = FAKE_VPNSERVICE_1
        self.svc_plugin.get_vpnservice_router_id.return_value = FAKE_ROUTER_ID
        self.driver = ipsec_driver.IPsecOvnVPNDriver(self.svc_plugin)
        self.driver.make_vpnservice_dict = mock.Mock(
            return_value={'id': FAKE_HOST})

    def test_get_vpn_services_on_host(self):
        with mock.patch('neutron_lib.context.get_admin_context') as get_ctx:
            mock_ctx = mock.Mock()
            mock_ctx.session.query().join().join().filter.return_value = [
                {'id': FAKE_HOST}]
            get_ctx.return_value = mock_ctx
            context = n_ctx.get_admin_context()
            driver_callback = ipsec_driver.IPsecVpnOvnDriverCallBack(
                self.driver)
            self.assertEqual(
                [{'id': FAKE_HOST}],
                driver_callback.get_vpn_services_on_host(
                    context, FAKE_HOST)
            )
            self.svc_plugin.get_vpn_agent_on_host.assert_called_once_with(
                mock.ANY, FAKE_HOST)
            self.svc_plugin.auto_schedule_routers.assert_called_once_with(
                mock.ANY, {'id': FAKE_HOST})
